"""Option discovery service for mapping options and identifying positional parameters."""

import inspect
from typing import Any

from freyja.utils.text_util import TextUtil

from .argument_preprocessor import PositionalInfo


class OptionDiscovery:
  """Discovers and categorizes all possible CLI options AND positional parameters by scope."""

  def __init__(self, command_tree, target_class: type | None = None):
    """Initialize option discovery service."""
    self.command_tree = command_tree
    self.target_class = target_class

  def discover_global_options(self) -> set[str]:
    """Discover global options from main class constructor."""
    if not self.target_class:
      return set()

    return self._analyze_constructor_params(self.target_class)

  def discover_subglobal_options(self) -> dict[str, set[str]]:
    """Discover sub-global options from inner class constructors."""
    subglobal_options = {}

    if not hasattr(self.command_tree, 'groups') or not self.command_tree.groups:
      return subglobal_options

    for group_name, group_info in self.command_tree.groups.items():
      if 'inner_class' in group_info and group_info['inner_class']:
        inner_class = group_info['inner_class']
        subglobal_options[group_name] = self._analyze_inner_class_params(inner_class)

    return subglobal_options

  def discover_command_options(self) -> dict[str, set[str]]:
    """Discover command-specific options from function signatures."""
    command_options = {}

    # Discover from flat commands
    if hasattr(self.command_tree, 'flat_commands') and self.command_tree.flat_commands:
      for cmd_name, cmd_info in self.command_tree.flat_commands.items():
        if 'function' in cmd_info:
          command_options[cmd_name] = self._analyze_function_params(cmd_info['function'])

    # Discover from grouped commands
    if hasattr(self.command_tree, 'groups') and self.command_tree.groups:
      for group_name, group_info in self.command_tree.groups.items():
        if 'commands' in group_info:
          for cmd_name, cmd_info in group_info['commands'].items():
            full_cmd_name = f"{group_name}--{cmd_name}"
            if 'function' in cmd_info:
              command_options[full_cmd_name] = self._analyze_function_params(cmd_info['function'])

    return command_options

  def discover_positional_parameters(self) -> dict[str, PositionalInfo]:
    """Discover positional parameters from function signatures."""
    positional_params = {}

    # Discover from flat commands
    if hasattr(self.command_tree, 'flat_commands') and self.command_tree.flat_commands:
      for cmd_name, cmd_info in self.command_tree.flat_commands.items():
        if 'function' in cmd_info:
          pos_info = self._find_positional_parameter(cmd_info['function'])
          if pos_info:
            positional_params[cmd_name] = pos_info

    # Discover from grouped commands
    if hasattr(self.command_tree, 'groups') and self.command_tree.groups:
      for group_name, group_info in self.command_tree.groups.items():
        if 'commands' in group_info:
          for cmd_name, cmd_info in group_info['commands'].items():
            full_cmd_name = f"{group_name}--{cmd_name}"
            if 'function' in cmd_info:
              pos_info = self._find_positional_parameter(cmd_info['function'])
              if pos_info:
                positional_params[full_cmd_name] = pos_info

    return positional_params

  def get_all_known_options(self, command_path: list[str] = None) -> set[str]:
    """Get all known options for a given command path."""
    all_options = set()

    # Add global options
    all_options.update(self.discover_global_options())

    # Add subglobal options for current path
    if command_path and len(command_path) > 0:
      subglobal_opts = self.discover_subglobal_options()
      group_name = command_path[0]
      if group_name in subglobal_opts:
        all_options.update(subglobal_opts[group_name])

    # Add command options for current command
    if command_path:
      command_opts = self.discover_command_options()
      if len(command_path) == 1:
        cmd_name = command_path[0]
      elif len(command_path) == 2:
        cmd_name = f"{command_path[0]}--{command_path[1]}"
      else:
        return all_options

      if cmd_name in command_opts:
        all_options.update(command_opts[cmd_name])

    # Add built-in options
    built_in_options = {'--help', '-h', '--no-color', '-n', '--verbose', '-v', '--_complete'}
    all_options.update(built_in_options)

    return all_options

  def _analyze_constructor_params(self, target_class: type) -> set[str]:
    """Analyze class constructor parameters to extract option names."""
    options = set()

    try:
      sig = inspect.signature(target_class.__init__)
      for param_name, param in sig.parameters.items():
        if param_name == 'self':
          continue
        flag_name = TextUtil.kebab_case(param_name)
        options.add(f"--{flag_name}")
    except (ValueError, TypeError):
      # Handle cases where signature inspection fails
      pass

    return options

  def _analyze_inner_class_params(self, inner_class: type) -> set[str]:
    """Analyze inner class constructor parameters to extract sub-global option names."""
    options = set()

    try:
      sig = inspect.signature(inner_class.__init__)
      params = list(sig.parameters.items())
      # Skip self (index 0) and main (index 1), start from index 2
      for param_name, param in params[2:]:
        flag_name = TextUtil.kebab_case(param_name)
        options.add(f"--{flag_name}")
    except (ValueError, TypeError):
      # Handle cases where signature inspection fails
      pass

    return options

  def _analyze_function_params(self, function: Any) -> set[str]:
    """Analyze function parameters to extract command option names."""
    options = set()

    if not function:
      return options

    try:
      sig = inspect.signature(function)
      for param_name, param in sig.parameters.items():
        if param_name == 'self':
          continue
        flag_name = TextUtil.kebab_case(param_name)
        options.add(f"--{flag_name}")
    except (ValueError, TypeError):
      # Handle cases where signature inspection fails
      pass

    return options

  def _find_positional_parameter(self, function: Any) -> PositionalInfo | None:
    """Find the first non-default parameter as positional parameter."""
    if not function:
      return None

    try:
      sig = inspect.signature(function)
      for param_name, param in sig.parameters.items():
        if param_name == 'self':
          continue
        # First parameter without default becomes positional
        if param.default == param.empty:
          return PositionalInfo(
            param_name=param_name,
            param_type=param.annotation if param.annotation != param.empty else str,
            is_required=True,
            help_text=f"{param_name} parameter"
          )
    except (ValueError, TypeError):
      # Handle cases where signature inspection fails
      pass

    return None

  def validate_option_conflicts(self) -> dict[str, list[str]]:
    """Detect option name conflicts across different scopes."""
    conflicts = {}

    global_options = self.discover_global_options()
    subglobal_options = self.discover_subglobal_options()
    command_options = self.discover_command_options()

    # Check for conflicts between global and subglobal
    for group_name, group_options in subglobal_options.items():
      overlapping = global_options.intersection(group_options)
      if overlapping:
        conflicts[f"global vs {group_name}"] = list(overlapping)

    # Check for conflicts between subglobal groups
    group_names = list(subglobal_options.keys())
    for i in range(len(group_names)):
      for j in range(i + 1, len(group_names)):
        group1, group2 = group_names[i], group_names[j]
        overlapping = subglobal_options[group1].intersection(subglobal_options[group2])
        if overlapping:
          conflicts[f"{group1} vs {group2}"] = list(overlapping)

    # Check for conflicts between command options in same scope
    command_conflicts = {}
    for cmd_name, cmd_options in command_options.items():
      for other_cmd, other_options in command_options.items():
        if cmd_name != other_cmd and cmd_name.split('--')[0] == other_cmd.split('--')[0]:
          # Same group, check for conflicts
          overlapping = cmd_options.intersection(other_options)
          if overlapping:
            pair_key = f"{cmd_name} vs {other_cmd}"
            if pair_key not in command_conflicts:
              command_conflicts[pair_key] = list(overlapping)

    conflicts.update(command_conflicts)

    return conflicts

  def suggest_option_corrections(self, unknown_option: str, command_path: list[str] = None) -> list[str]:
    """Suggest corrections for unknown options using similarity matching."""
    all_options = self.get_all_known_options(command_path)

    suggestions = []

    # Simple string similarity matching
    for option in all_options:
      if self._calculate_similarity(unknown_option, option) > 0.6:
        suggestions.append(option)

    # Sort by similarity score
    suggestions.sort(key=lambda x: self._calculate_similarity(unknown_option, x), reverse=True)

    return suggestions[:3]  # Return top 3 suggestions

  def _calculate_similarity(self, s1: str, s2: str) -> float:
    """Calculate simple string similarity score."""
    if not s1 or not s2:
      return 0.0

    # Remove leading dashes for comparison
    s1_clean = s1.lstrip('-')
    s2_clean = s2.lstrip('-')

    if s1_clean == s2_clean:
      return 1.0

    # Simple character overlap scoring
    s1_chars = set(s1_clean.lower())
    s2_chars = set(s2_clean.lower())

    if not s1_chars or not s2_chars:
      return 0.0

    intersection = s1_chars.intersection(s2_chars)
    union = s1_chars.union(s2_chars)

    return len(intersection) / len(union)



"""Integration test for validating links in markdown files."""
import re
import urllib.request
import urllib.error
from pathlib import Path
from typing import Any
import pytest


class TestMarkdownLinks:
  """Integration tests for markdown link validation."""

  @pytest.fixture
  def project_root(self) -> Path:
    """Get the project root directory."""
    return Path(__file__).parents[3]

  @pytest.fixture
  def markdown_files(self, project_root: Path) -> list[Path]:
    """Get all markdown files in the project."""
    markdown_files = []
    
    # Add root markdown files
    for pattern in ["*.md"]:
      markdown_files.extend(project_root.glob(pattern))
    
    # Add docs markdown files recursively
    docs_dir = project_root / "docs"
    if docs_dir.exists():
      markdown_files.extend(docs_dir.rglob("*.md"))
    
    return sorted(markdown_files)

  def extract_links(self, content: str, file_path: Path) -> list[tuple[str, int, str]]:
    """
    Extract all links from markdown content.
    
    :param content: Markdown file content
    :param file_path: Path to the markdown file for context
    :return: List of (link_url, line_number, link_type) tuples
    """
    links = []
    lines = content.split('\n')
    
    # Regex patterns for different link types
    patterns = [
      # Standard markdown links: [text](url)
      (r'\[([^\]]*)\]\(([^)]+)\)', 'markdown_link'),
      # Reference links: [text]: url
      (r'^\s*\[([^\]]*)\]:\s*(.+)$', 'reference_link'),
      # Image links: ![alt](url)
      (r'!\[([^\]]*)\]\(([^)]+)\)', 'image_link'),
      # HTML links: <a href="url">
      (r'<a[^>]+href=["\']([^"\']+)["\'][^>]*>', 'html_link'),
      # Plain URLs: http(s)://...
      (r'(?<!["\'\(])https?://[^\s<>\[\]]+(?!["\'\)])', 'plain_url'),
    ]
    
    for line_num, line in enumerate(lines, 1):
      for pattern, link_type in patterns:
        matches = re.finditer(pattern, line, re.IGNORECASE)
        for match in matches:
          if link_type == 'plain_url':
            url = match.group(0)
          else:
            url = match.group(2) if len(match.groups()) > 1 else match.group(1)
          
          # Skip empty links and anchors only
          if url and not url.startswith('#'):
            links.append((url.strip(), line_num, link_type))
    
    return links

  def is_external_url(self, url: str) -> bool:
    """Check if URL is external (http/https)."""
    return url.lower().startswith(('http://', 'https://'))

  def is_false_positive_url(self, url: str) -> bool:
    """Check if URL is a known false positive that shouldn't be tested."""
    false_positives = [
      # Git URLs with @ suffixes (not meant to be HTTP accessible)
      r'\.git@',
      # Malformed URLs from code snippets
      r'https?://["\')]',
      # URLs ending with quote characters (likely from code/examples)
      r'https?://[^"\']*["\')]$',
    ]
    
    import re
    for pattern in false_positives:
      if re.search(pattern, url):
        return True
    return False

  def is_valid_external_url(self, url: str) -> tuple[bool, str]:
    """
    Check if external URL is valid by making a HEAD request.
    
    :param url: URL to check
    :return: (is_valid, error_message)
    """
    if not self.is_external_url(url):
      return True, ""
    
    try:
      # Create request with User-Agent header to avoid blocking
      req = urllib.request.Request(url, method='HEAD')
      req.add_header('User-Agent', 'Mozilla/5.0 (compatible; Freyja-LinkChecker/1.0)')
      
      with urllib.request.urlopen(req, timeout=10) as response:
        # Accept any 2xx or 3xx status code
        return 200 <= response.status < 400, ""
        
    except urllib.error.HTTPError as e:
      # Some servers return 405 for HEAD requests, try GET with range
      if e.code == 405:
        try:
          req = urllib.request.Request(url)
          req.add_header('User-Agent', 'Mozilla/5.0 (compatible; Freyja-LinkChecker/1.0)')
          req.add_header('Range', 'bytes=0-1023')  # Only get first 1KB
          
          with urllib.request.urlopen(req, timeout=10) as response:
            return 200 <= response.status < 400, ""
        except Exception as inner_e:
          return False, f"HTTP {e.code}: {e.reason} (fallback failed: {inner_e})"
      return False, f"HTTP {e.code}: {e.reason}"
      
    except urllib.error.URLError as e:
      return False, f"URL Error: {e.reason}"
      
    except Exception as e:
      return False, f"Unexpected error: {e}"

  def resolve_relative_path(self, base_path: Path, relative_url: str) -> Path:
    """
    Resolve relative path from base file location.
    
    :param base_path: Base file path
    :param relative_url: Relative URL to resolve
    :return: Resolved absolute path
    """
    # Remove query strings and fragments
    clean_url = relative_url.split('?')[0].split('#')[0]
    
    # Resolve relative to the directory containing the markdown file
    base_dir = base_path.parent
    resolved = (base_dir / clean_url).resolve()
    
    return resolved

  def is_valid_relative_path(self, base_path: Path, relative_url: str) -> tuple[bool, str]:
    """
    Check if relative path exists.
    
    :param base_path: Base file path
    :param relative_url: Relative URL to check
    :return: (is_valid, error_message)
    """
    if self.is_external_url(relative_url):
      return True, ""
    
    try:
      resolved_path = self.resolve_relative_path(base_path, relative_url)
      
      if not resolved_path.exists():
        return False, f"File not found: {resolved_path}"
      
      # Get project root - find the root that contains both freyja/ and docs/
      project_root = base_path
      while project_root.parent != project_root:
        if (project_root / "freyja").exists() or (project_root / "pyproject.toml").exists():
          break
        project_root = project_root.parent
      
      # Check if it's within the project (but allow docs navigation)
      try:
        rel_path = resolved_path.relative_to(project_root)
        # Allow paths within the project structure
        return True, ""
      except ValueError:
        return False, f"Path outside project: {resolved_path}"
      
    except Exception as e:
      return False, f"Path resolution error: {e}"

  def test_markdown_files_exist(self, markdown_files: list[Path]) -> None:
    """Test that we found markdown files to validate."""
    assert len(markdown_files) > 0, "No markdown files found to validate"
    
    # Verify key files exist
    file_names = {f.name for f in markdown_files}
    assert "README.md" in file_names, "Root README.md not found"
    assert "CLAUDE.md" in file_names, "CLAUDE.md not found"

  def test_all_relative_links_valid(self, markdown_files: list[Path]) -> None:
    """Test that all relative links in markdown files point to valid locations."""
    failed_links = []
    
    # Get project root for relative path calculation
    project_root = markdown_files[0] if markdown_files else Path.cwd()
    while project_root.parent != project_root:
      if (project_root / "freyja").exists() or (project_root / "pyproject.toml").exists():
        break
      project_root = project_root.parent
    
    for md_file in markdown_files:
      content = md_file.read_text(encoding='utf-8')
      links = self.extract_links(content, md_file)
      
      for url, line_num, link_type in links:
        if not self.is_external_url(url):
          is_valid, error_msg = self.is_valid_relative_path(md_file, url)
          if not is_valid:
            try:
              relative_file_path = md_file.relative_to(project_root)
            except ValueError:
              relative_file_path = md_file.name
              
            failed_links.append({
              'file': str(relative_file_path),
              'line': line_num,
              'url': url,
              'type': link_type,
              'error': error_msg
            })
    
    if failed_links:
      error_msg = "Invalid relative links found:\n"
      for link in failed_links:
        error_msg += f"  {link['file']}:{link['line']} - {link['url']} ({link['type']}) - {link['error']}\n"
      
      pytest.fail(error_msg)

  @pytest.mark.slow
  def test_external_links_accessible(self, markdown_files: list[Path]) -> None:
    """Test that external links are accessible (marked as slow test)."""
    failed_links = []
    tested_urls = set()  # Avoid testing same URL multiple times
    
    for md_file in markdown_files:
      content = md_file.read_text(encoding='utf-8')
      links = self.extract_links(content, md_file)
      
      for url, line_num, link_type in links:
        if self.is_external_url(url) and url not in tested_urls and not self.is_false_positive_url(url):
          tested_urls.add(url)
          is_valid, error_msg = self.is_valid_external_url(url)
          if not is_valid:
            failed_links.append({
              'file': str(md_file.relative_to(md_file.parents[2] if 'tests' in str(md_file) else md_file.parents[0])),
              'line': line_num,
              'url': url,
              'type': link_type,
              'error': error_msg
            })
    
    if failed_links:
      error_msg = "Inaccessible external links found:\n"
      for link in failed_links:
        error_msg += f"  {link['file']}:{link['line']} - {link['url']} ({link['type']}) - {link['error']}\n"
      
      # Make this a warning instead of failure for external links
      # since external services may be temporarily unavailable
      import warnings
      warnings.warn(f"External link validation issues:\n{error_msg}")

  def test_no_broken_internal_references(self, markdown_files: list[Path]) -> None:
    """Test that internal markdown references (anchors) are valid."""
    failed_refs = []
    
    # Get project root for relative path calculation
    project_root = markdown_files[0] if markdown_files else Path.cwd()
    while project_root.parent != project_root:
      if (project_root / "freyja").exists() or (project_root / "pyproject.toml").exists():
        break
      project_root = project_root.parent
    
    for md_file in markdown_files:
      content = md_file.read_text(encoding='utf-8')
      
      # Find all internal references (links with #)
      internal_links = re.findall(r'\[([^\]]*)\]\(([^)]*#[^)]+)\)', content)
      
      for link_text, full_url in internal_links:
        if '#' in full_url:
          url_part, anchor = full_url.split('#', 1)
          
          # If url_part is empty, it's a same-file anchor
          if not url_part:
            target_file = md_file
          else:
            if self.is_external_url(url_part):
              continue  # Skip external anchors
            
            # Resolve relative path
            target_file = self.resolve_relative_path(md_file, url_part)
            if not target_file.exists():
              try:
                relative_file_path = md_file.relative_to(project_root)
              except ValueError:
                relative_file_path = md_file.name
                
              failed_refs.append({
                'file': str(relative_file_path),
                'anchor': full_url,
                'error': f'Target file not found: {target_file}'
              })
              continue
          
          # Check if anchor exists in target file
          target_content = target_file.read_text(encoding='utf-8')
          
          # Generate possible anchor formats
          anchor_variants = [
            anchor.lower().replace(' ', '-').replace('_', '-'),
            anchor.lower().replace(' ', '').replace('_', ''),
            anchor.replace(' ', '-').replace('_', '-'),
          ]
          
          # Look for headers that could match this anchor
          headers = re.findall(r'^#+\s+(.+)$', target_content, re.MULTILINE)
          header_anchors = []
          for header in headers:
            # Strip markdown formatting and convert to anchor format
            clean_header = re.sub(r'[^\w\s-]', '', header).lower()
            header_anchors.extend([
              clean_header.replace(' ', '-').replace('_', '-'),
              clean_header.replace(' ', '').replace('_', ''),
            ])
          
          # Check if any variant matches
          if not any(variant in header_anchors or variant in target_content.lower() for variant in anchor_variants):
            try:
              relative_file_path = md_file.relative_to(project_root)
            except ValueError:
              relative_file_path = md_file.name
              
            failed_refs.append({
              'file': str(relative_file_path),
              'anchor': full_url,
              'error': f'Anchor #{anchor} not found in {target_file.name}'
            })
    
    if failed_refs:
      error_msg = "Broken internal references found:\n"
      for ref in failed_refs:
        error_msg += f"  {ref['file']} - {ref['anchor']} - {ref['error']}\n"
      
      pytest.fail(error_msg)

  def test_link_extraction_accuracy(self, project_root: Path) -> None:
    """Test that link extraction works correctly with sample content."""
    sample_content = """
# Test Document

[Standard link](./docs/README.md)
[External link](https://github.com/terracoil/freyja)
![Image](./images/logo.png)
[Reference link][ref1]
<a href="./manual-link.html">HTML link</a>
Plain URL: https://pypi.org/project/freyja/

[ref1]: ./referenced.md
"""
    
    test_file = project_root / "test.md"
    links = self.extract_links(sample_content, test_file)
    
    # Should extract all links
    assert len(links) >= 6, f"Expected at least 6 links, got {len(links)}"
    
    # Check specific link types are found
    link_urls = [url for url, _, _ in links]
    assert "./docs/README.md" in link_urls
    assert "https://github.com/terracoil/freyja" in link_urls
    assert "./images/logo.png" in link_urls
    assert "https://pypi.org/project/freyja/" in link_urls


"""Fish shell completion handler."""


from .base import CompletionContext, CompletionHandler


class FishCompletionHandler(CompletionHandler):
  """Fish-specific completion handler."""

  def generate_script(self, prog_name: str) -> str:
    """Generate fish completion script."""
    script = f'''# Fish completion for {prog_name}
# Generated by freyja

function __{prog_name}_complete
    # Set up completion environment
    set -x _FREYJA_COMPLETE fish
    set -x COMP_WORDS_STR (commandline -cp)
    set -x COMP_CWORD_NUM (count (commandline -cp))

    # Get completions from the program
    {prog_name} --_complete 2>/dev/null
end

# Register completions for {prog_name}
complete -f -c {prog_name} -a '(__{prog_name}_complete)'

# Add option completions
complete -c {prog_name} -l help -d "Show help message"
complete -c {prog_name} -l verbose -d "Enable verbose output"  
complete -c {prog_name} -l no-color -d "Disable colored output"
complete -c {prog_name} -l install-completion -d "Install shell completion"
'''
    return script

  def get_completions(self, context: CompletionContext) -> list[str]:
    """Get fish-specific completions."""
    # Reuse bash completion logic for now
    from .bash import BashCompletionHandler
    bash_handler = BashCompletionHandler(self.cli)
    return bash_handler._get_generic_completions(context)

  def install_completion(self, prog_name: str) -> bool:
    """Install fish completion."""
    from .installer import CompletionInstaller
    installer = CompletionInstaller(self, prog_name)
    return installer.install('fish')



"""Tests for System class and inner classes."""

import sys
from unittest.mock import MagicMock, patch

import pytest

from freyja import FreyjaCLI
from freyja.cli import SystemClassBuilder

System = SystemClassBuilder.build(True, True)

class TestSystem:
  """Test the System class initialization."""

  def test_system_initialization(self):
    """Test System class can be initialized with default config dir."""
    system = System()
    assert system is not None

class TestSystemTuneTheme:
  """Test the System.TuneTheme inner class."""

  def test_tune_theme_initialization_universal(self):
    """Test TuneTheme initializes with universal theme by default."""
    tuner = System().TuneTheme()
    assert tuner.use_colorful_theme is False
    assert tuner.adjust_percent == 0.0
    assert tuner.ADJUSTMENT_INCREMENT == 0.05

  def test_tune_theme_initialization_colorful(self):
    """Test TuneTheme initializes with colorful theme when specified."""
    tuner = System().TuneTheme(initial_theme="colorful")
    assert tuner.use_colorful_theme is True

  def test_increase_adjustment(self):
    """Test adjustment increase functionality."""
    tuner = System().TuneTheme()
    initial_adjustment = tuner.adjust_percent
    tuner.increase_adjustment()
    assert tuner.adjust_percent == initial_adjustment + tuner.ADJUSTMENT_INCREMENT

  def test_decrease_adjustment(self):
    """Test adjustment decrease functionality."""
    tuner = System().TuneTheme()
    tuner.adjust_percent = 0.10  # Set initial value
    tuner.decrease_adjustment()
    assert tuner.adjust_percent == 0.05

  def test_decrease_adjustment_minimum_bound(self):
    """Test adjustment decrease respects minimum bound."""
    tuner = System().TuneTheme()
    tuner.adjust_percent = -5.0  # Set at minimum
    tuner.decrease_adjustment()
    assert tuner.adjust_percent == -5.0  # Should not go lower

  def test_increase_adjustment_maximum_bound(self):
    """Test adjustment increase respects maximum bound."""
    tuner = System().TuneTheme()
    tuner.adjust_percent = 5.0  # Set at maximum
    tuner.increase_adjustment()
    assert tuner.adjust_percent == 5.0  # Should not go higher

  def test_toggle_theme(self):
    """Test theme toggling functionality."""
    tuner = System().TuneTheme()
    assert tuner.use_colorful_theme is False
    tuner.toggle_theme()
    assert tuner.use_colorful_theme is True
    tuner.toggle_theme()
    assert tuner.use_colorful_theme is False

  def test_select_strategy_with_valid_string(self):
    """Test strategy selection with valid string."""
    tuner = System().TuneTheme()
    tuner.select_strategy("COLOR_HSL")
    # Should set strategy without error
    from freyja.theme import AdjustStrategy
    assert tuner.adjust_strategy == AdjustStrategy.COLOR_HSL

  def test_select_strategy_with_invalid_string(self):
    """Test strategy selection with invalid string."""
    tuner = System().TuneTheme()
    original_strategy = tuner.adjust_strategy
    tuner.select_strategy("INVALID_STRATEGY")
    # Should not change strategy
    assert tuner.adjust_strategy == original_strategy

  def test_get_current_theme(self):
    """Test getting current theme with adjustments."""
    tuner = System().TuneTheme()
    theme = tuner.get_current_theme()
    assert theme is not None
    # Should return a theme object
    assert hasattr(theme, 'title')
    assert hasattr(theme, 'command_name')

  def test_individual_color_overrides(self):
    """Test individual color override functionality."""
    tuner = System().TuneTheme()
    assert len(tuner.individual_color_overrides) == 0

    # Add a color override
    from freyja.theme import RGB
    test_color = RGB.from_rgb(0xFF0000)  # Red
    tuner.individual_color_overrides['title'] = test_color
    tuner.modified_components.add('title')

    assert len(tuner.individual_color_overrides) == 1
    assert 'title' in tuner.modified_components

    # Get theme with overrides
    theme = tuner.get_current_theme()
    assert theme.title.fg == test_color

  def test_reset_component_color(self):
    """Test resetting individual component colors."""
    tuner = System().TuneTheme()
    from freyja.theme import RGB
    test_color = RGB.from_rgb(0xFF0000)

    # Add override then reset
    tuner.individual_color_overrides['title'] = test_color
    tuner.modified_components.add('title')
    tuner._reset_component_color('title')

    assert 'title' not in tuner.individual_color_overrides
    assert 'title' not in tuner.modified_components

  def test_reset_all_individual_colors(self):
    """Test resetting all individual colors."""
    tuner = System().TuneTheme()
    from freyja.theme import RGB

    # Add multiple overrides
    tuner.individual_color_overrides['title'] = RGB.from_rgb(0xFF0000)
    tuner.individual_color_overrides['command_name'] = RGB.from_rgb(0x00FF00)
    tuner.modified_components.update(['title', 'command_name'])

    tuner._reset_all_individual_colors()

    assert len(tuner.individual_color_overrides) == 0
    assert len(tuner.modified_components) == 0


class TestSystemCompletion:
  """Test the System.Completion inner class."""

  def test_completion_initialization(self):
    """Test Completion class initializes correctly."""
    completion = System().Completion()
    # Shell should auto-detect (could be bash, zsh, fish, etc.)
    assert completion.shell in ['bash', 'zsh', 'fish', 'powershell']
    assert completion._cli_instance is None
    assert completion._completion_handler is None

  def test_completion_initialization_with_cli(self):
    """Test Completion class initializes with FreyjaCLI instance."""
    mock_cli = MagicMock()
    completion = System.Completion(cli_instance=mock_cli)
    assert completion._cli_instance == mock_cli

  def test_completion_explicit_shell(self):
    """Test Completion class with explicitly specified shell."""
    completion = System().Completion(shell='bash')
    assert completion.shell == 'bash'

    completion = System().Completion(shell='zsh')
    assert completion.shell == 'zsh'

  def test_is_completion_request_false(self):
    """Test completion request detection returns False normally."""
    completion = System().Completion()
    # Patch sys.argv to not include completion flags
    with patch.object(sys, 'argv', ['script.py']):
      with patch('os.environ.get', return_value=None):
        assert completion.is_completion_request() is False

  def test_is_completion_request_true_argv(self):
    """Test completion request detection returns True with --_complete flag."""
    completion = System().Completion()
    with patch.object(sys, 'argv', ['script.py', '--_complete']):
      assert completion.is_completion_request() is True

  def test_is_completion_request_true_env(self):
    """Test completion request detection returns True with env var."""
    completion = System().Completion()
    with patch.object(sys, 'argv', ['script.py']):
      with patch('os.environ.get', return_value='1'):
        assert completion.is_completion_request() is True

  def test_install_completion_disabled(self):
    """Test install completion when completion is disabled."""
    mock_cli = MagicMock()
    mock_cli.enable_completion = False
    completion = System.Completion(cli_instance=mock_cli)

    result = completion.install()
    assert result is False

  def test_show_completion_disabled(self):
    """Test show completion when completion is disabled."""
    mock_cli = MagicMock()
    mock_cli.enable_completion = False
    completion = System.Completion(cli_instance=mock_cli)

    # Should not raise error, just print message
    completion.show("bash")

  def test_install_completion_no_cli_instance(self):
    """Test install completion without FreyjaCLI instance."""
    completion = System().Completion()
    result = completion.install()
    assert result is False

  @patch('sys.exit')
  def test_handle_completion_no_handler(self, mock_exit):
    """Test handle completion without completion handler."""
    completion = System().Completion()
    # Mock the init_completion to prevent actually initializing
    with patch.object(completion, 'init_completion'):
      completion.handle_completion()
      # Should call exit once with code 1 when no handler is available
      mock_exit.assert_called_once_with(1)


class TestSystemCLIGeneration:
  """Test System class FreyjaCLI generation."""

  def test_system_cli_generation(self):
    """Test System class generates FreyjaCLI correctly."""
    cli = FreyjaCLI(System)
    parser = cli.create_parser()

    # Should have System class as target
    assert cli.target_class == System
    assert cli.target_mode.value == 'class'

  def test_system_cli_has_inner_classes(self):
    """Test System FreyjaCLI recognizes inner classes."""
    cli = FreyjaCLI(System)

    # Should discover inner classes as command groups
    assert 'tune-theme' in cli.commands
    assert cli.commands['tune-theme']['type'] == 'group'
    assert 'completion' in cli.commands
    assert cli.commands['completion']['type'] == 'group'

  def test_system_cli_command_structure(self):
    """Test System FreyjaCLI creates proper command structure."""
    cli = FreyjaCLI(System)

    # Should have hierarchical cmd_tree
    assert 'tune-theme' in cli.commands
    assert 'completion' in cli.commands

    # Groups should have hierarchical structure
    tune_theme_group = cli.commands['tune-theme']
    assert tune_theme_group['type'] == 'group'
    assert 'increase-adjustment' in tune_theme_group['cmd_tree']
    assert 'decrease-adjustment' in tune_theme_group['cmd_tree']

    completion_group = cli.commands['completion']
    assert completion_group['type'] == 'group'
    assert 'install' in completion_group['cmd_tree']
    assert 'show' in completion_group['cmd_tree']

  def test_system_tune_theme_methods(self):
    """Test System FreyjaCLI includes TuneTheme methods as hierarchical command groups."""
    cli = FreyjaCLI(System)

    # Check that TuneTheme methods are included as cmd_tree under tune-theme group
    tune_theme_group = cli.commands['tune-theme']
    expected_commands = [
      'increase-adjustment', 'decrease-adjustment',
      'select-strategy', 'toggle-theme',
      'edit-colors', 'show-rgb', 'run-interactive'
    ]

    for command in expected_commands:
      assert command in tune_theme_group['cmd_tree']
      assert tune_theme_group['cmd_tree'][command]['type'] == 'command'

  def test_system_completion_methods(self):
    """Test System FreyjaCLI includes Completion methods as hierarchical command groups."""
    cli = FreyjaCLI(System)

    # Check that Completion methods are included as cmd_tree under completion group
    completion_group = cli.commands['completion']
    expected_commands = ['install', 'show']

    for command in expected_commands:
      assert command in completion_group['cmd_tree']
      assert completion_group['cmd_tree'][command]['type'] == 'command'

  def test_system_cli_execution(self):
    """Test System FreyjaCLI can execute cmd_tree."""
    cli = FreyjaCLI(System)

    # Test that we can create a parser without errors
    parser = cli.create_parser()
    assert parser is not None

    # Test parsing a valid hierarchical command
    args = parser.parse_args(['tune-theme', 'toggle-theme'])
    assert hasattr(args, '_cli_function')
    assert args._cli_function is not None


class TestSystemIntegration:
  """Integration tests for System class with FreyjaCLI."""

  def test_system_help_generation(self):
    """Test System FreyjaCLI generates help correctly."""
    cli = FreyjaCLI(System, title="System Test FreyjaCLI")
    parser = cli.create_parser()

    help_text = parser.format_help()

    # Should contain main sections
    assert "System Test FreyjaCLI" in help_text
    assert "tune-theme" in help_text
    assert "completion" in help_text

  def test_system_command_group_help(self):
    """Test System FreyjaCLI command group help generation."""
    cli = FreyjaCLI(System)
    parser = cli.create_parser()

    # Test that parsing to command level works (help would exit)
    with pytest.raises(SystemExit):
      parser.parse_args(['tune-theme', '--help'])

if __name__ == '__main__':
  pytest.main([__file__])



"""Argument preprocessor for flexible option ordering and positional parameter support."""

import inspect
from dataclasses import dataclass
from typing import Any

from freyja.command.command_tree import CommandTree
from freyja.parser.docstring_parser import DocStringParser
from freyja.utils.text_util import TextUtil


@dataclass
class PositionalInfo:
  """Information about a positional parameter."""
  param_name: str
  param_type: type
  is_required: bool
  help_text: str | None = None


class ArgumentPreprocessor:
  """Preprocesses command line arguments to enable flexible option ordering AND positional parameter support."""

  def __init__(self, command_tree: CommandTree, target_class: type | None = None):
    """Initialize the preprocessor with command tree and target class info."""
    self.command_tree = command_tree
    self.target_class = target_class
    self._global_options: set[str] = set()
    self._subglobal_options: dict[str, set[str]] = {}  # group_name -> set of options
    self._command_options: dict[str, set[str]] = {}    # command_name -> set of options
    self._positional_params: dict[str, PositionalInfo] = {}  # command_name -> positional parameter info
    self._build_option_maps()
    self._build_positional_maps()

  def validate_arguments(self, args: list[str]) -> tuple[bool, list[str]]:
    """Validate arguments before preprocessing."""
    errors = []

    # Basic validation - check for obviously malformed arguments
    for i, arg in enumerate(args):
      if arg.startswith('--') and '=' in arg:
        # Handle --flag=value format
        flag_part = arg.split('=')[0]
        if not self._is_known_option(flag_part):
          errors.append(f"Unknown option: {flag_part}")
      elif arg.startswith('--') and not self._is_known_option(arg):
        # Check if next arg exists and isn't a flag (for --flag value format)
        if i + 1 < len(args) and not args[i + 1].startswith('-'):
          continue  # This is likely --flag value format, will be validated later
        errors.append(f"Unknown option: {arg}")

    return len(errors) == 0, errors

  def preprocess_args(self, args: list[str]) -> list[str]:
    """Reorder arguments to match argparse hierarchical expectations."""
    # Parse command structure from positional arguments
    command_path = self._extract_command_path(args)

    # Convert flag-style positional parameters to positional arguments for backward compatibility
    args_with_converted_positionals = self._convert_positional_flags_to_positional(args, command_path)

    # Categorize options by scope
    categorized_args = self._categorize_arguments(args_with_converted_positionals, command_path)

    # Reorder arguments: global -> group -> command -> remaining positionals
    return self._reorder_arguments(categorized_args, command_path)

  def _build_option_maps(self) -> None:
    """Build maps of which options belong to which scope."""
    # Build global options from target class constructor
    if self.target_class:
      self._global_options = self._extract_constructor_options(self.target_class)
    else:
      # Fallback global options
      self._global_options = {'-n', '--no-color', '-h', '--help'}

    # Build sub-global options from inner classes and command options
    if self.command_tree:
      for group_name, group_info in self.command_tree.tree.items():
        if group_info.get('type') == 'group':
          # This is a group, check for inner_class
          if group_info.get('inner_class'):
            self._subglobal_options[group_name] = self._extract_constructor_options(group_info['inner_class'], skip_params=1)

          # Process commands within the group
          if 'cmd_tree' in group_info:
            for cmd_name, cmd_info in group_info['cmd_tree'].items():
              if cmd_info.get('type') == 'command' and cmd_info.get('function'):
                full_cmd_name = f"{group_name}--{cmd_name}"  # Use double-dash format
                self._command_options[full_cmd_name] = self._extract_function_options(cmd_info['function'])
        elif group_info.get('type') == 'command':
          # This is a flat command
          if group_info.get('function'):
            self._command_options[group_name] = self._extract_function_options(group_info['function'])

  def _build_positional_maps(self) -> None:
    """Build maps of positional parameters for each command."""
    if not self.command_tree:
      return

    for group_name, group_info in self.command_tree.tree.items():
      if group_info.get('type') == 'group':
        # Process commands within the group
        if 'cmd_tree' in group_info:
          for cmd_name, cmd_info in group_info['cmd_tree'].items():
            if cmd_info.get('type') == 'command' and cmd_info.get('function'):
              full_cmd_name = f"{group_name}--{cmd_name}"  # Use double-dash format
              positional_info = self._extract_positional_parameter(cmd_info['function'])
              if positional_info:
                self._positional_params[full_cmd_name] = positional_info
      elif group_info.get('type') == 'command':
        # This is a flat command
        if group_info.get('function'):
          positional_info = self._extract_positional_parameter(group_info['function'])
          if positional_info:
            self._positional_params[group_name] = positional_info

  def _extract_constructor_options(self, cls: type, skip_params: int = 1) -> set[str]:
    """Extract options from class constructor."""
    options = set()
    if not hasattr(cls, '__init__'):
      return options

    sig = inspect.signature(cls.__init__)
    params = list(sig.parameters.items())

    # Skip 'self' and potentially other params (like 'main' in inner classes)
    for param_name, param in params[skip_params:]:
      if param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
        continue

      flag_name = f"--{TextUtil.kebab_case(param_name)}"
      options.add(flag_name)

    return options

  def _extract_function_options(self, func: Any) -> set[str]:
    """Extract options from function signature."""
    options = set()
    sig = inspect.signature(func)

    for param_name, param in sig.parameters.items():
      if param_name == 'self' or param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
        continue

      # Skip positional parameters (first param without default)
      if self._is_first_non_default_param(func, param_name):
        continue

      flag_name = f"--{TextUtil.kebab_case(param_name)}"
      options.add(flag_name)

    return options

  def _extract_positional_parameter(self, func: Any) -> PositionalInfo | None:
    """Extract positional parameter info from function signature."""
    sig = inspect.signature(func)
    _, param_help = DocStringParser.extract_function_help(func)

    for param_name, param in sig.parameters.items():
      if param_name == 'self' or param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
        continue

      # First parameter without default value becomes positional
      if param.default == param.empty:
        return PositionalInfo(
          param_name=param_name,
          param_type=param.annotation if param.annotation != param.empty else str,
          is_required=True,
          help_text=param_help.get(param_name)
        )

    return None

  def _is_first_non_default_param(self, func: Any, param_name: str) -> bool:
    """Check if this is the first parameter without a default value."""
    sig = inspect.signature(func)

    for name, param in sig.parameters.items():
      if name == 'self' or param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
        continue

      if param.default == param.empty:
        return name == param_name  # This is the first non-default param

    return False

  def _extract_command_path(self, args: list[str]) -> list[str]:
    """Extract hierarchical command path from arguments."""
    command_path = []

    for arg in args:
      if arg.startswith('-'):
        break  # Stop at first option

      # Check if this arg is a valid command/group name
      if self._is_command_or_group(command_path + [arg]):
        command_path.append(arg)
      else:
        # This might be a positional parameter, stop building command path
        break

    return command_path

  def _is_command_or_group(self, path: list[str]) -> bool:
    """Check if the given path represents a valid command or group."""
    if not self.command_tree or not path:
      return False

    # Check if it's a group
    if len(path) == 1 and path[0] in self.command_tree.tree:
      group_info = self.command_tree.tree[path[0]]
      if group_info.get('type') == 'group':
        return True

    # Check if it's a command (full path)
    if len(path) >= 2:
      command_name = '--'.join(path)  # Inner class commands use double-dash
      # Check in the command tree structure
      group_name = path[0]
      if group_name in self.command_tree.tree:
        group_info = self.command_tree.tree[group_name]
        if group_info.get('type') == 'group' and 'cmd_tree' in group_info:
          cmd_name = path[1]
          if cmd_name in group_info['cmd_tree']:
            return True

    # Check if it's a flat command
    if len(path) == 1 and path[0] in self.command_tree.tree:
      cmd_info = self.command_tree.tree[path[0]]
      if cmd_info.get('type') == 'command':
        return True

    return False

  def _convert_positional_flags_to_positional(self, args: list[str], command_path: list[str]) -> list[str]:
    """Convert flag-style positional parameters back to positional for backward compatibility.

    Example: ['function-with-types', '--text', 'hello'] -> ['function-with-types', 'hello']
    """
    if not command_path:
      return args

    # Get the appropriate command name
    command_name = '--'.join(command_path) if len(command_path) > 1 else command_path[0] if command_path else None
    if not command_name:
      return args

    # Check if this command has a positional parameter
    positional_info = self._positional_params.get(command_name)
    if not positional_info:
      return args

    # Look for the flag version of the positional parameter
    flag_name = f"--{TextUtil.kebab_case(positional_info.param_name)}"
    result_args = args.copy()

    # Find flag and its value, convert to positional
    i = 0
    while i < len(result_args) - 1:  # -1 because we need to check next element
      if result_args[i] == flag_name:
        # Found the flag, next element should be the value
        if i + 1 < len(result_args) and not result_args[i + 1].startswith('-'):
          flag_value = result_args[i + 1]
          # Remove the flag and its value
          result_args.pop(i)  # Remove flag
          result_args.pop(i)  # Remove value (index shifts after first pop)
          # Find where to insert the positional value (after command path)
          insert_position = len(command_path)
          result_args.insert(insert_position, flag_value)
          break
      i += 1

    return result_args

  def _handle_positional_parameters(self, args: list[str], command_path: list[str]) -> list[str]:
    """Convert positional parameters to flag format for argparse compatibility."""
    if not command_path:
      return args

    # Get command name (for flat commands or hierarchical commands)
    if len(command_path) == 1:
      command_name = command_path[0]
    else:
      command_name = '--'.join(command_path)

    # Check if this command has a positional parameter
    positional_info = self._positional_params.get(command_name)
    if not positional_info:
      return args

    # Find the positional value in the args
    result_args = args.copy()
    command_path_end = len(command_path)
    flag_name = f"--{TextUtil.kebab_case(positional_info.param_name)}"

    # Look for positional value after command path
    i = command_path_end
    while i < len(args):
      arg = args[i]

      # Skip options and their values
      if arg.startswith('-'):
        # Check if this is the flag we're looking for
        if arg == flag_name:
          # The flag already exists, no need to convert positional
          return args

        # Skip this option and its value if it exists
        i += 1
        if i < len(args) and not args[i].startswith('-'):
          i += 1  # Skip the option value
        continue

      # This is a potential positional value
      # Replace positional value with flag format
      result_args[i:i+1] = [flag_name, arg]
      break

    return result_args

  def _categorize_arguments(self, args: list[str], command_path: list[str]) -> dict[str, list[str]]:
    """Categorize arguments by scope (global, sub-global, command)."""
    categorized = {
      'global': [],
      'subglobal': [],
      'command': [],
      'positional': command_path.copy()
    }

    i = 0
    # Skip command path
    while i < len(args) and i < len(command_path):
      if args[i] == command_path[i]:
        i += 1
      else:
        break

    # Categorize remaining arguments
    while i < len(args):
      arg = args[i]

      if not arg.startswith('-'):
        # Non-option argument, probably a value
        categorized['command'].append(arg)
        i += 1
        continue

      # This is an option, determine its scope
      if arg in self._global_options:
        categorized['global'].append(arg)
        # Check if it has a value
        if i + 1 < len(args) and not args[i + 1].startswith('-'):
          categorized['global'].append(args[i + 1])
          i += 2
        else:
          i += 1
      elif self._is_subglobal_option(arg, command_path):
        categorized['subglobal'].append(arg)
        # Check if it has a value
        if i + 1 < len(args) and not args[i + 1].startswith('-'):
          categorized['subglobal'].append(args[i + 1])
          i += 2
        else:
          i += 1
      else:
        # Assume it's a command option
        categorized['command'].append(arg)
        # Check if it has a value
        if i + 1 < len(args) and not args[i + 1].startswith('-'):
          categorized['command'].append(args[i + 1])
          i += 2
        else:
          i += 1

    return categorized

  def _is_subglobal_option(self, option: str, command_path: list[str]) -> bool:
    """Check if an option is a sub-global option for the current command path."""
    if not command_path:
      return False

    group_name = command_path[0]  # First element is the group
    return group_name in self._subglobal_options and option in self._subglobal_options[group_name]

  def _reorder_arguments(self, categorized: dict[str, list[str]], command_path: list[str]) -> list[str]:
    """Reorder arguments into hierarchical format."""
    reordered = []

    # Add global options first
    reordered.extend(categorized['global'])

    # Add command path
    reordered.extend(categorized['positional'])

    # Add sub-global options
    reordered.extend(categorized['subglobal'])

    # Add command-specific options
    reordered.extend(categorized['command'])

    return reordered

  def _is_known_option(self, option: str) -> bool:
    """Check if an option is known in any scope."""
    if option in self._global_options:
      return True

    for group_options in self._subglobal_options.values():
      if option in group_options:
        return True

    for command_options in self._command_options.values():
      if option in command_options:
        return True

    # Check for built-in options
    builtin_options = {'--help', '-h', '--no-color', '-n'}
    if option in builtin_options:
      return True

    return False



"""System-level FreyjaCLI cmd_tree for freyja."""

import os

from freyja.theme import (
  RGB,
  AdjustStrategy,
  ColorFormatter,
  create_default_theme,
  create_default_theme_colorful,
)
from freyja.theme.theme_style import ThemeStyle
from freyja.utils.ansi_string import AnsiString


class TuneTheme:
  """Interactive theme tuning utility."""

  # Adjustment increment constant for easy modification
  ADJUSTMENT_INCREMENT = 0.05

  def __init__(self, initial_theme: str = "universal"):
    """Initialize theme tuner.

    :param initial_theme: Starting theme (universal or colorful)
    """
    self.adjust_percent = 0.0
    self.adjust_strategy = AdjustStrategy.LINEAR
    self.use_colorful_theme = initial_theme.lower() == "colorful"
    self.formatter = ColorFormatter(enable_colors=True)

    # Individual color override tracking
    self.individual_color_overrides: dict[str, RGB] = {}
    self.modified_components: set[str] = set()

    # Theme component metadata for user interface
    self.theme_components = [
      ("title", "Title text"),
      ("subtitle", "Section headers (COMMANDS:, OPTIONS:)"),
      ("command_name", "Command names"),
      ("command_description", "Command descriptions"),
      # Command Group Level (inner class level)
      ("command_group_name", "Command group names (inner class names)"),
      ("command_group_description", "Command group descriptions (inner class descriptions)"),
      ("command_group_option_name", "Command group option flags"),
      ("command_group_option_description", "Command group option descriptions"),
      # Grouped Command Level (cmd_tree within the group)
      ("grouped_command_name", "Grouped command names (methods within groups)"),
      ("grouped_command_description", "Grouped command descriptions (method descriptions)"),
      ("grouped_command_option_name", "Grouped command option flags"),
      ("grouped_command_option_description", "Grouped command option descriptions"),
      ("option_name", "Regular option flags (--name)"),
      ("option_description", "Regular option descriptions"),
      ("required_asterisk", "Required field markers (*)")
    ]

    # Get terminal width
    try:
      self.console_width = os.get_terminal_size().columns
    except (OSError, ValueError):
      self.console_width = int(os.environ.get('COLUMNS', 80))

  def increase_adjustment(self) -> None:
    """Increase color adjustment by 0.05."""
    self.adjust_percent = min(5.0, self.adjust_percent + self.ADJUSTMENT_INCREMENT)
    print(f"Adjustment increased to {self.adjust_percent:.2f}")

  def decrease_adjustment(self) -> None:
    """Decrease color adjustment by 0.05."""
    self.adjust_percent = max(-5.0, self.adjust_percent - self.ADJUSTMENT_INCREMENT)
    print(f"Adjustment decreased to {self.adjust_percent:.2f}")

  def select_strategy(self, strategy: str = None) -> None:
    """Select color adjustment strategy.

    :param strategy: Strategy name or None for interactive selection
    """
    if strategy:
      # Convert string to enum if valid
      try:
        self.adjust_strategy = AdjustStrategy[strategy.upper()]
        print(f"Strategy set to: {self.adjust_strategy.name}")
      except KeyError:
        print(f"Invalid strategy: {strategy}")
    else:
      # Interactive selection
      self._select_adjustment_strategy()

  def toggle_theme(self) -> None:
    """Toggle between universal and colorful themes."""
    self.use_colorful_theme = not self.use_colorful_theme
    theme_name = "COLORFUL" if self.use_colorful_theme else "UNIVERSAL"
    print(f"Theme toggled to {theme_name}")

  def edit_colors(self) -> None:
    """Edit individual color values interactively."""
    self.edit_individual_color()

  def show_rgb(self) -> None:
    """Display RGB values for current theme colors."""
    self.display_rgb_values()
    input("\nPress Enter to continue...")

  def run_interactive(self) -> None:
    """Run the interactive theme tuner."""
    self.run_interactive_menu()

  def get_current_theme(self):
    """Get theme with global adjustments and individual overrides applied."""
    # 1. Start with base theme
    base_theme = create_default_theme_colorful() if self.use_colorful_theme else create_default_theme()

    # 2. Apply global adjustments if any
    if self.adjust_percent != 0.0:
      try:
        adjusted_theme = base_theme.create_adjusted_copy(
          adjust_percent=self.adjust_percent,
          adjust_strategy=self.adjust_strategy
        )
      except ValueError:
        adjusted_theme = base_theme
    else:
      adjusted_theme = base_theme

    # 3. Apply individual color overrides if any
    result = adjusted_theme
    if self.individual_color_overrides:
      result = self._apply_individual_overrides(adjusted_theme)

    return result

  def _apply_individual_overrides(self, theme):
    """Create new theme with individual color overrides applied."""
    from freyja.theme.theme import Theme

    # Get all current theme styles
    theme_styles = {}
    for component_name, _ in self.theme_components:
      original_style = getattr(theme, component_name)

      if component_name in self.individual_color_overrides:
        # Create new ThemeStyle with overridden color but preserve other attributes
        override_color = self.individual_color_overrides[component_name]
        theme_styles[component_name] = ThemeStyle(
          fg=override_color,
          bg=original_style.bg,
          bold=original_style.bold,
          italic=original_style.italic,
          dim=original_style.dim,
          underline=original_style.underline
        )
      else:
        # Use original style
        theme_styles[component_name] = original_style

    # Create new theme with overridden styles
    return Theme(
      adjust_percent=theme.adjust_percent,
      adjust_strategy=theme.adjust_strategy,
      **theme_styles
    )

  def display_theme_info(self):
    """Display current theme information and preview."""
    theme = self.get_current_theme()

    # Create a fresh formatter with the current adjusted theme
    current_formatter = ColorFormatter(enable_colors=True)

    # Simple header
    print("=" * min(self.console_width, 60))
    print("🎛️  THEME TUNER")
    print("=" * min(self.console_width, 60))

    # Current settings
    strategy_name = self.adjust_strategy.name
    theme_name = "COLORFUL" if self.use_colorful_theme else "UNIVERSAL"

    print(f"Theme: {theme_name}")
    print(f"Strategy: {strategy_name}")
    print(f"Adjust: {self.adjust_percent:.2f}")

    # Show modification status
    if self.individual_color_overrides:
      modified_count = len(self.individual_color_overrides)
      total_count = len(self.theme_components)
      modified_names = ', '.join(sorted(self.individual_color_overrides.keys()))
      print(f"Modified Components: {modified_count}/{total_count} ({modified_names})")
    else:
      print("Modified Components: None")
    print()

    # Simple preview with real-time color updates
    print("📋 FreyjaCLI Preview:")
    print(
      f"  {current_formatter.apply_style('hello', theme.command_name)}: {current_formatter.apply_style('Greet the user', theme.command_description)}"
    )
    print(
      f"  {current_formatter.apply_style('--name NAME', theme.option_name)}: {current_formatter.apply_style('Specify name', theme.option_description)}"
    )
    print(
      f"  {current_formatter.apply_style('--email EMAIL', theme.option_name)} {current_formatter.apply_style('*', theme.required_asterisk)}: {current_formatter.apply_style('Required email', theme.option_description)}"
    )
    print()

  def display_rgb_values(self):
    """Display RGB values for theme incorporation with names colored in their RGB values on different backgrounds."""
    theme = self.get_current_theme()  # Get the current adjusted theme

    print("\n" + "=" * min(self.console_width, 60))
    print("🎨 RGB VALUES FOR THEME INCORPORATION")
    print("=" * min(self.console_width, 60))

    # Color mappings for the current adjusted theme - include ALL theme components
    color_map = [
      ("title", theme.title.fg, "Title color"),
      ("subtitle", theme.subtitle.fg, "Subtitle color"),
      ("command_name", theme.command_name.fg, "Command name"),
      ("command_description", theme.command_description.fg, "Command description"),
      # Command Group Level (inner class level)
      ("command_group_name", theme.command_group_name.fg, "Command group name"),
      ("command_group_description", theme.command_group_description.fg, "Command group description"),
      ("command_group_option_name", theme.command_group_option_name.fg, "Command group option name"),
      ("command_group_option_description", theme.command_group_option_description.fg, "Command group option description"),
      # Grouped Command Level (cmd_tree within the group)
      ("grouped_command_name", theme.grouped_command_name.fg, "Grouped command name"),
      ("grouped_command_description", theme.grouped_command_description.fg, "Grouped command description"),
      ("grouped_command_option_name", theme.grouped_command_option_name.fg, "Grouped command option name"),
      ("grouped_command_option_description", theme.grouped_command_option_description.fg, "Grouped command option description"),
      ("option_name", theme.option_name.fg, "Option name"),
      ("option_description", theme.option_description.fg, "Option description"),
      ("required_asterisk", theme.required_asterisk.fg, "Required asterisk"),
    ]

    # Create background colors for testing readability
    white_bg = RGB.from_rgb(0xFFFFFF)  # White background
    black_bg = RGB.from_rgb(0x000000)  # Black background

    # Collect theme code components
    theme_code_lines = []

    for name, color_code, description in color_map:
      if isinstance(color_code, RGB):
        # Check if this component has been modified
        is_modified = name in self.individual_color_overrides

        # RGB instance - show name in the actual color
        r, g, b = color_code.to_ints()
        hex_code = color_code.to_hex()
        hex_int = f"0x{hex_code[1:]}"  # Convert #FF80FF to 0xFF80FF

        # Get the complete theme style for this component (includes bold, italic, etc.)
        current_theme_style = getattr(theme, name)

        # Create styled versions using the complete theme style with different backgrounds
        # Only the white/black background versions should be styled
        white_bg_style = ThemeStyle(
          fg=color_code,
          bg=white_bg,
          bold=current_theme_style.bold,
          italic=current_theme_style.italic,
          dim=current_theme_style.dim,
          underline=current_theme_style.underline
        )
        black_bg_style = ThemeStyle(
          fg=color_code,
          bg=black_bg,
          bold=current_theme_style.bold,
          italic=current_theme_style.italic,
          dim=current_theme_style.dim,
          underline=current_theme_style.underline
        )

        # Apply styles (first name is unstyled, only white/black background versions are styled)
        colored_name_white = self.formatter.apply_style(name, white_bg_style)
        colored_name_black = self.formatter.apply_style(name, black_bg_style)

        # First name display is just plain text with standard padding
        padding = 20 - len(name)
        padded_name = name + ' ' * padding

        # Show modification indicator
        modifier_indicator = " [CUSTOM]" if is_modified else ""

        print(f"  {padded_name} = rgb({r:3}, {g:3}, {b:3})  # {hex_code}{modifier_indicator}")

        # Show original color if modified
        if is_modified:
          # Get the original color (before override)
          base_theme = create_default_theme_colorful() if self.use_colorful_theme else create_default_theme()
          adjusted_base = base_theme
          if self.adjust_percent != 0.0:
            try:
              adjusted_base = base_theme.create_adjusted_copy(
                adjust_percent=self.adjust_percent,
                adjust_strategy=self.adjust_strategy
              )
            except ValueError:
              adjusted_base = base_theme

          original_style = getattr(adjusted_base, name)
          if original_style.fg and isinstance(original_style.fg, RGB):
            orig_r, orig_g, orig_b = original_style.fg.to_ints()
            orig_hex = original_style.fg.to_hex()
            print(f"    Original: rgb({orig_r:3}, {orig_g:3}, {orig_b:3})  # {orig_hex}")

        # Calculate alignment width based on longest component name for clean f-string alignment
        max_component_name_length = max(len(comp_name) for comp_name, _ in self.theme_components)
        white_field_width = max_component_name_length + 2  # +2 for spacing buffer

        # Use AnsiString for proper f-string alignment with ANSI escape codes
        print(
          f"    On white: {AnsiString(colored_name_white):<{white_field_width}}On black: {AnsiString(colored_name_black)}")
        print()

        # Build theme code line for this color
        # Handle background colors and text styles
        additional_styles = []
        if hasattr(theme, name):
          style_obj = getattr(theme, name)
          if style_obj.bg:
            if isinstance(style_obj.bg, RGB):
              bg_r, bg_g, bg_b = style_obj.bg.to_ints()
              bg_hex = style_obj.bg.to_hex()
              bg_hex_int = f"0x{bg_hex[1:]}"
              additional_styles.append(f"bg=RGB.from_rgb({bg_hex_int})")
          if style_obj.bold:
            additional_styles.append("bold=True")
          if style_obj.italic:
            additional_styles.append("italic=True")
          if style_obj.dim:
            additional_styles.append("dim=True")
          if style_obj.underline:
            additional_styles.append("underline=True")

        # Create ThemeStyle constructor call
        style_params = [f"fg=RGB.from_rgb({hex_int})"]
        style_params.extend(additional_styles)
        style_call = f"ThemeStyle({', '.join(style_params)})"

        theme_code_lines.append(f"    {name}={style_call},")

      elif color_code and isinstance(color_code, str) and color_code.startswith('#'):
        # Hex string (fallback handling)
        try:
          hex_clean = color_code.strip().lstrip('#').upper()
          if len(hex_clean) == 3:
            hex_clean = ''.join(c * 2 for c in hex_clean)
          if len(hex_clean) == 6 and all(c in '0123456789ABCDEF' for c in hex_clean):
            hex_int = int(hex_clean, 16)
            rgb = RGB.from_rgb(hex_int)
            r, g, b = rgb.to_ints()

            padding = 20 - len(name)
            padded_name = name + ' ' * padding

            print(f"  {padded_name} = rgb({r:3}, {g:3}, {b:3})  # {color_code}")

            # Add to theme code
            hex_int_str = f"0x{hex_clean}"
            theme_code_lines.append(f"    {name}=ThemeStyle(fg=RGB.from_rgb({hex_int_str})),")
          else:
            print(f"  {name:20} = {color_code}")
        except ValueError:
          print(f"  {name:20} = {color_code}")
      elif color_code:
        print(f"  {name:20} = {color_code}")
      else:
        # No color defined
        print(f"  {name:20} = (no color)")

    # Display the complete theme creation code
    print("\n" + "=" * min(self.console_width, 60))
    print("📋 THEME CREATION CODE")
    print("=" * min(self.console_width, 60))
    print()
    print("from freyja.theme import RGB, ThemeStyle, Theme")
    print()
    print("def create_custom_theme() -> Theme:")
    print("  \"\"\"Create a custom theme with the current colors.\"\"\"")
    print("  return Theme(")

    for line in theme_code_lines:
      print(line)

    print("  )")
    print()
    print("# Usage in your FreyjaCLI:")
    print("from freyja.cli import FreyjaCLI")
    print("cli = FreyjaCLI(your_module, theme=create_custom_theme())")
    print("cli.display()")

    print("=" * min(self.console_width, 60))

  def edit_individual_color(self):
    """Interactive color editing for individual theme components."""
    while True:
      print("\n" + "=" * min(self.console_width, 60))
      print("🎨 EDIT INDIVIDUAL COLOR")
      print("=" * min(self.console_width, 60))

      # Display components with modification indicators
      for i, (component_name, description) in enumerate(self.theme_components, 1):
        is_modified = component_name in self.individual_color_overrides
        status = " [MODIFIED]" if is_modified else ""
        print(f"  {i:2d}. {component_name:<25} {status}")
        print(f"      {description}")

        # Show current color
        current_theme = self.get_current_theme()
        current_style = getattr(current_theme, component_name)
        if current_style.fg and isinstance(current_style.fg, RGB):
          hex_color = current_style.fg.to_hex()
          r, g, b = current_style.fg.to_ints()
          colored_preview = self.formatter.apply_style("██", ThemeStyle(fg=current_style.fg))
          print(f"      Current: {colored_preview} rgb({r:3}, {g:3}, {b:3}) {hex_color}")
        print()

      print("Commands:")
      print("  Enter number (1-12) to edit component")
      print("  [x] Reset all individual colors")
      print("  [q] Return to main menu")

      try:
        choice = input("\nChoice: ").lower().strip()

        if choice == 'q':
          break
        elif choice == 'x':
          self._reset_all_individual_colors()
          print("All individual color overrides reset!")
          continue

        # Try to parse as component number
        try:
          component_index = int(choice) - 1
          if 0 <= component_index < len(self.theme_components):
            component_name, description = self.theme_components[component_index]
            self._edit_component_color(component_name, description)
          else:
            print(f"Invalid choice. Please enter 1-{len(self.theme_components)}")
        except ValueError:
          print("Invalid input. Please enter a number or command.")

      except (KeyboardInterrupt, EOFError):
        break

  def _edit_component_color(self, component_name: str, description: str):
    """Edit color for a specific component."""
    # Get current color
    current_theme = self.get_current_theme()
    current_style = getattr(current_theme, component_name)
    current_color = current_style.fg if current_style.fg else RGB.from_rgb(0x808080)

    is_modified = component_name in self.individual_color_overrides

    print(f"\n🎨 EDITING: {component_name}")
    print(f"Description: {description}")

    if isinstance(current_color, RGB):
      hex_color = current_color.to_hex()
      r, g, b = current_color.to_ints()
      colored_preview = self.formatter.apply_style("██████", ThemeStyle(fg=current_color))
      print(f"Current: {colored_preview} rgb({r:3}, {g:3}, {b:3}) {hex_color}")

    if is_modified:
      print("(This color has been customized)")

    print("\nInput Methods:")
    print("  [h] Hex color entry (e.g., FF8080)")
    print("  [r] Reset to original color")
    print("  [q] Cancel")

    try:
      method = input("\nChoose input method: ").lower().strip()

      if method == 'q':
        return
      elif method == 'r':
        self._reset_component_color(component_name)
        print(f"Reset {component_name} to original color!")
        return
      elif method == 'h':
        self._hex_color_input(component_name, current_color)
      else:
        print("Invalid choice.")

    except (KeyboardInterrupt, EOFError):
      return

  def _hex_color_input(self, component_name: str, current_color: RGB):
    """Handle hex color input for a component."""
    print(f"\nCurrent color: {current_color.to_hex()}")
    print("Enter new hex color (without #):")
    print("Examples: FF8080, ff8080, F80 (short form)")

    try:
      hex_input = input("Hex color: ").strip()

      if not hex_input:
        print("No input provided, canceling.")
        return

      # Normalize hex input
      hex_clean = hex_input.upper().lstrip('#')

      # Handle 3-character hex (e.g., F80 -> FF8800)
      if len(hex_clean) == 3:
        hex_clean = ''.join(c * 2 for c in hex_clean)

      # Validate hex
      if len(hex_clean) != 6 or not all(c in '0123456789ABCDEF' for c in hex_clean):
        print("Invalid hex color format. Please use 6 digits (e.g., FF8080)")
        return

      # Convert to RGB
      hex_int = int(hex_clean, 16)
      new_color = RGB.from_rgb(hex_int)

      # Preview the new color
      r, g, b = new_color.to_ints()
      colored_preview = self.formatter.apply_style("██████", ThemeStyle(fg=new_color))
      print(f"\nPreview: {colored_preview} rgb({r:3}, {g:3}, {b:3}) #{hex_clean}")

      # Confirm
      confirm = input("Apply this color? [y/N]: ").lower().strip()
      if confirm in ('y', 'yes'):
        self.individual_color_overrides[component_name] = new_color
        self.modified_components.add(component_name)
        print(f"✅ Applied new color to {component_name}!")
      else:
        print("Color change canceled.")

    except (KeyboardInterrupt, EOFError):
      print("\nColor editing canceled.")
    except ValueError as e:
      print(f"Error: {e}")

  def _reset_component_color(self, component_name: str):
    """Reset a component's color to original."""
    if component_name in self.individual_color_overrides:
      del self.individual_color_overrides[component_name]
      self.modified_components.discard(component_name)

  def _reset_all_individual_colors(self):
    """Reset all individual color overrides."""
    self.individual_color_overrides.clear()
    self.modified_components.clear()

  def _select_adjustment_strategy(self):
    """Allow user to select from all available adjustment strategies."""
    strategies = list(AdjustStrategy)

    print("\n🎯 SELECT ADJUSTMENT STRATEGY")
    print("=" * 40)

    # Display current strategy
    current_index = strategies.index(self.adjust_strategy)
    print(f"Current strategy: {self.adjust_strategy.name}")
    print()

    # Display all available strategies with numbers
    print("Available strategies:")
    strategy_descriptions = {
      AdjustStrategy.LINEAR: "Linear blend adjustment (legacy compatibility)",
      AdjustStrategy.COLOR_HSL: "HSL-based lightness adjustment",
      AdjustStrategy.MULTIPLICATIVE: "Simple RGB value scaling",
      AdjustStrategy.GAMMA: "Gamma correction for perceptual uniformity",
      AdjustStrategy.LUMINANCE: "ITU-R BT.709 perceived brightness adjustment",
      AdjustStrategy.OVERLAY: "Photoshop-style overlay blend mode",
      AdjustStrategy.ABSOLUTE: "Legacy absolute color adjustment"
    }

    for i, strategy in enumerate(strategies, 1):
      marker = "→" if strategy == self.adjust_strategy else " "
      description = strategy_descriptions.get(strategy, "Color adjustment strategy")
      print(f"{marker} [{i}] {strategy.name}: {description}")

    print()
    print("  [Enter] Keep current strategy")
    print("  [q] Cancel")

    try:
      choice = input("\nSelect strategy (1-7): ").strip().lower()

      if choice == '' or choice == 'q':
        return  # Keep current strategy

      try:
        strategy_index = int(choice) - 1
        if 0 <= strategy_index < len(strategies):
          old_strategy = self.adjust_strategy.name
          self.adjust_strategy = strategies[strategy_index]
          print(f"✅ Strategy changed from {old_strategy} to {self.adjust_strategy.name}")
        else:
          print("❌ Invalid strategy number. Strategy unchanged.")
      except ValueError:
        print("❌ Invalid input. Strategy unchanged.")

    except (EOFError, KeyboardInterrupt):
      print("\n❌ Selection cancelled.")

  def run_interactive_menu(self):
    """Run a simple menu-based theme tuner."""
    print("🎛️  THEME TUNER")
    print("=" * 40)
    print("Interactive controls are not available in this environment.")
    print("Using simple menu mode instead.")
    print()

    while True:
      self.display_theme_info()

      print("Available cmd_tree:")
      print(f"  [+] Increase adjustment by {self.ADJUSTMENT_INCREMENT}")
      print(f"  [-] Decrease adjustment by {self.ADJUSTMENT_INCREMENT}")
      print("  [s] Select adjustment strategy")
      print("  [t] Toggle theme (universal/colorful)")
      print("  [e] Edit individual colors")
      print("  [r] Show RGB values")
      print("  [q] Quit")

      try:
        choice = input("\nEnter command: ").lower().strip()

        if choice == 'q':
          break
        elif choice == '+':
          self.increase_adjustment()
        elif choice == '-':
          self.decrease_adjustment()
        elif choice == 's':
          self._select_adjustment_strategy()
        elif choice == 't':
          self.toggle_theme()
        elif choice == 'e':
          self.edit_colors()
        elif choice == 'r':
          self.show_rgb()
        else:
          print("Invalid command. Try again.")

        print()

      except (KeyboardInterrupt, EOFError):
        break

    print("\n🎨 Theme tuning session ended.")



"""Tests for ArgumentPreprocessor class."""
import inspect
from unittest.mock import Mock

import pytest

from freyja.command.command_info import CommandInfo
from freyja.command.command_tree import CommandTree
from freyja.parser.argument_preprocessor import ArgumentPreprocessor, PositionalInfo


class TestArgumentPreprocessor:
  """Test ArgumentPreprocessor class functionality."""

  @pytest.fixture
  def simple_command_tree(self):
    """Create a simple command tree for testing."""
    tree = CommandTree()

    # Add a simple function mock
    mock_func = Mock()
    mock_func.__name__ = "test_function"

    # Create CommandInfo for the mock function
    command_info = CommandInfo(
      name="test-command",
      original_name="test_function",
      function=mock_func,
      signature=inspect.signature(lambda: None),
      docstring="Test command"
    )

    tree.add_command("test-command", command_info)

    return tree

  @pytest.fixture
  def hierarchical_command_tree(self):
    """Create hierarchical command tree for testing."""
    tree = CommandTree()

    # Mock inner class with constructor
    class MockInnerClass:
      def __init__(self, database_url: str = "sqlite:///test.db", timeout: int = 30):
        self.database_url = database_url
        self.timeout = timeout

    # Add group with inner class (this should provide the sub-global options)
    tree.add_group("data-ops", "Data operations group", MockInnerClass)

    # Mock function for hierarchical command
    mock_func = Mock()
    mock_func.__name__ = "process_data"

    # Create CommandInfo for the hierarchical command
    command_info = CommandInfo(
      name="process",
      original_name="process_data",
      function=mock_func,
      signature=inspect.signature(lambda: None),
      docstring="Process data"
    )

    # Add command directly to the group (not subgroup for this test)
    tree.add_command_to_group("data-ops", "process", command_info)

    return tree

  @pytest.fixture
  def target_class_with_params(self):
    """Create target class with constructor parameters."""
    class TestClass:
      def __init__(self, config_file: str = "config.json", debug: bool = False):
        self.config_file = config_file
        self.debug = debug

    return TestClass

  def test_preprocessor_initialization_simple(self, simple_command_tree):
    """Test ArgumentPreprocessor initialization with simple command tree."""
    preprocessor = ArgumentPreprocessor(simple_command_tree)

    assert preprocessor.command_tree == simple_command_tree
    assert preprocessor.target_class is None
    assert isinstance(preprocessor._global_options, set)
    assert isinstance(preprocessor._command_options, dict)
    assert isinstance(preprocessor._positional_params, dict)

  def test_preprocessor_initialization_with_class(self, simple_command_tree, target_class_with_params):
    """Test ArgumentPreprocessor initialization with target class."""
    preprocessor = ArgumentPreprocessor(simple_command_tree, target_class_with_params)

    assert preprocessor.target_class == target_class_with_params
    assert "--config-file" in preprocessor._global_options
    assert "--debug" in preprocessor._global_options

  def test_global_options_extraction(self, simple_command_tree, target_class_with_params):
    """Test extraction of global options from target class constructor."""
    preprocessor = ArgumentPreprocessor(simple_command_tree, target_class_with_params)

    # Should extract options from constructor
    assert "--config-file" in preprocessor._global_options
    assert "--debug" in preprocessor._global_options


  def test_command_path_extraction_simple(self, simple_command_tree):
    """Test command path extraction for simple commands."""
    preprocessor = ArgumentPreprocessor(simple_command_tree)

    args = ["test-command", "--option", "value"]
    command_path = preprocessor._extract_command_path(args)

    assert command_path == ["test-command"]

  def test_command_path_extraction_hierarchical(self, hierarchical_command_tree):
    """Test command path extraction for hierarchical commands."""
    preprocessor = ArgumentPreprocessor(hierarchical_command_tree)

    args = ["data-ops", "process", "--option", "value"]
    command_path = preprocessor._extract_command_path(args)

    assert command_path == ["data-ops", "process"]

  def test_command_path_extraction_with_options(self, simple_command_tree):
    """Test command path extraction stops at first option."""
    preprocessor = ArgumentPreprocessor(simple_command_tree)

    args = ["test-command", "--option", "value", "more-args"]
    command_path = preprocessor._extract_command_path(args)

    assert command_path == ["test-command"]

  def test_is_known_option_global(self, simple_command_tree, target_class_with_params):
    """Test recognition of global options."""
    preprocessor = ArgumentPreprocessor(simple_command_tree, target_class_with_params)

    assert preprocessor._is_known_option("--config-file")
    assert preprocessor._is_known_option("--debug")
    assert preprocessor._is_known_option("--help")
    assert not preprocessor._is_known_option("--unknown-option")

  def test_is_known_option_builtin(self, simple_command_tree):
    """Test recognition of built-in options."""
    preprocessor = ArgumentPreprocessor(simple_command_tree)

    assert preprocessor._is_known_option("--help")
    assert preprocessor._is_known_option("-h")
    assert preprocessor._is_known_option("--no-color")
    assert preprocessor._is_known_option("-n")

  def test_validate_arguments_valid(self, simple_command_tree, target_class_with_params):
    """Test argument validation with valid arguments."""
    preprocessor = ArgumentPreprocessor(simple_command_tree, target_class_with_params)

    args = ["test-command", "--config-file", "config.json", "--debug"]
    is_valid, errors = preprocessor.validate_arguments(args)

    assert is_valid
    assert len(errors) == 0

  def test_validate_arguments_invalid(self, simple_command_tree):
    """Test argument validation with invalid arguments."""
    preprocessor = ArgumentPreprocessor(simple_command_tree)

    args = ["test-command", "--unknown-flag", "--another-unknown"]
    is_valid, errors = preprocessor.validate_arguments(args)

    assert not is_valid
    assert len(errors) > 0
    assert any("unknown-flag" in error for error in errors)

  def test_validate_arguments_flag_value_format(self, simple_command_tree, target_class_with_params):
    """Test validation handles --flag=value format."""
    preprocessor = ArgumentPreprocessor(simple_command_tree, target_class_with_params)

    args = ["test-command", "--config-file=config.json"]
    is_valid, errors = preprocessor.validate_arguments(args)

    assert is_valid
    assert len(errors) == 0

  def test_preprocess_args_simple_command(self, simple_command_tree):
    """Test argument preprocessing for simple command."""
    preprocessor = ArgumentPreprocessor(simple_command_tree)

    args = ["test-command", "--verbose", "--help"]
    processed = preprocessor.preprocess_args(args)

    # Should have command followed by options in proper order
    assert "test-command" in processed
    assert "--verbose" in processed or "-v" in processed

  def test_preprocess_args_reordering(self, simple_command_tree, target_class_with_params):
    """Test that global options are reordered to come first."""
    preprocessor = ArgumentPreprocessor(simple_command_tree, target_class_with_params)

    # Global options mixed with command
    args = ["test-command", "--debug", "--config-file", "test.json"]
    processed = preprocessor.preprocess_args(args)

    # Global options should be moved to front
    debug_idx = processed.index("--debug")
    config_idx = processed.index("--config-file")
    command_idx = processed.index("test-command")

    assert debug_idx < command_idx
    assert config_idx < command_idx

  def test_subglobal_options_extraction(self, hierarchical_command_tree):
    """Test extraction of sub-global options from inner classes."""
    preprocessor = ArgumentPreprocessor(hierarchical_command_tree)

    # Should extract sub-global options from inner class constructor
    assert "data-ops" in preprocessor._subglobal_options
    subglobal_opts = preprocessor._subglobal_options["data-ops"]
    assert "--database-url" in subglobal_opts
    assert "--timeout" in subglobal_opts

  def test_is_subglobal_option(self, hierarchical_command_tree):
    """Test identification of sub-global options."""
    preprocessor = ArgumentPreprocessor(hierarchical_command_tree)

    command_path = ["data-ops", "process"]
    assert preprocessor._is_subglobal_option("--database-url", command_path)
    assert preprocessor._is_subglobal_option("--timeout", command_path)
    assert not preprocessor._is_subglobal_option("--unknown-option", command_path)

  def test_categorize_arguments_global(self, simple_command_tree, target_class_with_params):
    """Test argument categorization for global options."""
    preprocessor = ArgumentPreprocessor(simple_command_tree, target_class_with_params)

    args = ["test-command", "--config-file", "config.json", "--debug"]
    command_path = ["test-command"]
    categorized = preprocessor._categorize_arguments(args, command_path)

    assert "--config-file" in categorized['global']
    assert "config.json" in categorized['global']
    assert "--debug" in categorized['global']
    assert categorized['positional'] == command_path

  def test_categorize_arguments_mixed(self, hierarchical_command_tree, target_class_with_params):
    """Test argument categorization with mixed option types."""
    preprocessor = ArgumentPreprocessor(hierarchical_command_tree, target_class_with_params)

    args = [
      "data-ops", "process",
      "--config-file", "global.json",  # global
      "--database-url", "sqlite://test.db",  # sub-global
      "--some-command-option", "value"  # command
    ]
    command_path = ["data-ops", "process"]
    categorized = preprocessor._categorize_arguments(args, command_path)

    assert "--config-file" in categorized['global']
    assert "--database-url" in categorized['subglobal']
    assert "--some-command-option" in categorized['command']

  def test_reorder_arguments(self, simple_command_tree, target_class_with_params):
    """Test argument reordering maintains proper hierarchy."""
    preprocessor = ArgumentPreprocessor(simple_command_tree, target_class_with_params)

    categorized = {
      'global': ['--config-file', 'config.json'],
      'subglobal': ['--database-url', 'sqlite://test.db'],
      'command': ['--option', 'value'],
      'positional': ['test-command']
    }
    command_path = ['test-command']

    reordered = preprocessor._reorder_arguments(categorized, command_path)

    # Check order: global -> positional -> sub-global -> command
    config_idx = reordered.index('--config-file')
    command_idx = reordered.index('test-command')
    option_idx = reordered.index('--option')

    assert config_idx < command_idx < option_idx


class TestPositionalInfo:
  """Test PositionalInfo dataclass."""

  def test_positional_info_creation(self):
    """Test creation of PositionalInfo."""
    info = PositionalInfo("test_param", str, True, "Test parameter help")

    assert info.param_name == "test_param"
    assert info.param_type == str
    assert info.is_required is True
    assert info.help_text == "Test parameter help"

  def test_positional_info_without_help(self):
    """Test creation of PositionalInfo without help text."""
    info = PositionalInfo("test_param", int, False)

    assert info.param_name == "test_param"
    assert info.param_type == int
    assert info.is_required is False
    assert info.help_text is None


class TestPositionalParameterExtraction:
  """Test positional parameter extraction functionality."""

  def test_extract_positional_parameter_with_required(self):
    """Test extraction of positional parameter from function with required param."""
    def test_func(required_param: str, optional_param: str = "default"):
      return f"{required_param} {optional_param}"

    tree = CommandTree()
    preprocessor = ArgumentPreprocessor(tree)

    positional_info = preprocessor._extract_positional_parameter(test_func)

    assert positional_info is not None
    assert positional_info.param_name == "required_param"
    assert positional_info.param_type == str
    assert positional_info.is_required is True

  def test_extract_positional_parameter_no_required(self):
    """Test extraction when no required parameters exist."""
    def test_func(optional_param: str = "default", another_optional: int = 42):
      return f"{optional_param} {another_optional}"

    tree = CommandTree()
    preprocessor = ArgumentPreprocessor(tree)

    positional_info = preprocessor._extract_positional_parameter(test_func)

    assert positional_info is None

  def test_extract_positional_parameter_skip_self(self):
    """Test extraction skips 'self' parameter for methods."""
    class TestClass:
      def test_method(self, required_param: str, optional_param: str = "default"):
        return f"{required_param} {optional_param}"

    tree = CommandTree()
    preprocessor = ArgumentPreprocessor(tree)

    positional_info = preprocessor._extract_positional_parameter(TestClass.test_method)

    assert positional_info is not None
    assert positional_info.param_name == "required_param"

  def test_is_first_non_default_param(self):
    """Test identification of first non-default parameter."""
    def test_func(first_required: str, second_required: int, optional: str = "default"):
      return f"{first_required} {second_required} {optional}"

    tree = CommandTree()
    preprocessor = ArgumentPreprocessor(tree)

    assert preprocessor._is_first_non_default_param(test_func, "first_required") is True
    assert preprocessor._is_first_non_default_param(test_func, "second_required") is False
    assert preprocessor._is_first_non_default_param(test_func, "optional") is False

  def test_is_first_non_default_param_with_self(self):
    """Test identification with self parameter."""
    class TestClass:
      def test_method(self, first_required: str, optional: str = "default"):
        return f"{first_required} {optional}"

    tree = CommandTree()
    preprocessor = ArgumentPreprocessor(tree)

    assert preprocessor._is_first_non_default_param(TestClass.test_method, "first_required") is True
    assert preprocessor._is_first_non_default_param(TestClass.test_method, "optional") is False


class TestEdgeCases:
  """Test edge cases for ArgumentPreprocessor."""

  def test_empty_command_tree(self):
    """Test preprocessor with empty command tree."""
    tree = CommandTree()
    preprocessor = ArgumentPreprocessor(tree)

    args = ["--help"]
    processed = preprocessor.preprocess_args(args)

    # Should handle gracefully
    assert isinstance(processed, list)

  def test_no_target_class(self):
    """Test preprocessor without target class."""
    tree = CommandTree()
    preprocessor = ArgumentPreprocessor(tree, None)

    # Should use fallback global options (no longer includes verbose)
    assert '--no-color' in preprocessor._global_options or '-n' in preprocessor._global_options
    assert '--help' in preprocessor._global_options or '-h' in preprocessor._global_options

  def test_malformed_arguments(self):
    """Test handling of malformed arguments."""
    tree = CommandTree()
    preprocessor = ArgumentPreprocessor(tree)

    # Test with mixed valid and invalid args
    args = ["--help", "command", "--unknown", "value"]
    processed = preprocessor.preprocess_args(args)

    # Should process without crashing
    assert isinstance(processed, list)

  def test_option_without_value(self):
    """Test handling of boolean options without values."""
    tree = CommandTree()
    preprocessor = ArgumentPreprocessor(tree)

    args = ["command", "--help", "--verbose"]
    processed = preprocessor.preprocess_args(args)

    # Should handle boolean flags correctly
    assert "--help" in processed or "-h" in processed



from enum import Enum


class Fore(Enum):
  """Foreground color constants."""
  BLACK = 0x000000
  RED = 0xFF0000
  GREEN = 0x008000
  YELLOW = 0xFFFF00
  BLUE = 0x0000FF
  MAGENTA = 0xFF00FF
  CYAN = 0x00FFFF
  WHITE = 0xFFFFFF

  # Bright colors
  LIGHTBLACK_EX = 0x808080
  LIGHTRED_EX = 0xFF8080
  LIGHTGREEN_EX = 0x80FF80
  LIGHTYELLOW_EX = 0xFFFF80
  LIGHTBLUE_EX = 0x8080FF
  LIGHTMAGENTA_EX = 0xFF80FF
  LIGHTCYAN_EX = 0x80FFFF
  LIGHTWHITE_EX = 0xF0F0F0


class Back(Enum):
  """Background color constants."""
  BLACK = 0x000000
  RED = 0xFF0000
  GREEN = 0x008000
  YELLOW = 0xFFFF00
  BLUE = 0x0000FF
  MAGENTA = 0xFF00FF
  CYAN = 0x00FFFF
  WHITE = 0xFFFFFF

  # Bright backgrounds
  LIGHTBLACK_EX = 0x808080
  LIGHTRED_EX = 0xFF8080
  LIGHTGREEN_EX = 0x80FF80
  LIGHTYELLOW_EX = 0xFFFF80
  LIGHTBLUE_EX = 0x8080FF
  LIGHTMAGENTA_EX = 0xFF80FF
  LIGHTCYAN_EX = 0x80FFFF
  LIGHTWHITE_EX = 0xF0F0F0


class Style(Enum):
  """Text style constants."""
  DIM = '\x1b[2m'
  RESET_ALL = '\x1b[0m'
  # All ANSI style codes in one place
  ANSI_BOLD = '\x1b[1m'  # Bold text
  ANSI_ITALIC = '\x1b[3m'  # Italic text (support varies by terminal)
  ANSI_UNDERLINE = '\x1b[4m'  # Underlined text


class ForeUniversal(Enum):
  """Universal foreground color palette with carefully curated colors."""

  # Blues
  BLUE = 0x2196F3  # Material Blue 500
  OKABE_BLUE = 0x0072B2  # Okabe-Ito Blue
  INDIGO = 0x3F51B5  # Material Indigo 500
  SKY_BLUE = 0x56B4E9  # Sky Blue

  # Greens
  BLUISH_GREEN = 0x009E73  # Bluish Green
  GREEN = 0x4CAF50  # Material Green 500
  DARK_GREEN = 0x08780D  # Dark green
  TEAL = 0x009688  # Material Teal 500

  # Orange/Yellow
  ORANGE = 0xE69F00  # Okabe-Ito Orange
  MATERIAL_ORANGE = 0xFF9800  # Material Orange 500
  GOLD = 0xF39C12  # Muted Gold

  # Red/Magenta
  VERMILION = 0xD55E00  # Okabe-Ito Vermilion
  REDDISH_PURPLE = 0xCC79A7  # Reddish Purple

  # Purple
  PURPLE = 0x9C27B0  # Material Purple 500
  DEEP_PURPLE = 0x673AB7  # Material Deep Purple 500

  # Neutrals
  BLUE_GREY = 0x607D8B  # Material Blue Grey 500
  DARK_BROWN = 0x604030  # Material Brown 500
  BROWN = 0x795548  # Material Brown 500
  MEDIUM_GREY = 0x757575  # Medium Grey
  IBM_GREY = 0x8D8D8D  # IBM Gray 50
  LIGHT_GREY = 0xCCCCCC  # Light grey



"""Tests for HierarchicalHelpFormatter functionality."""

import argparse
from unittest.mock import Mock, patch

from freyja.help.help_formatter import HierarchicalHelpFormatter
from freyja.theme import create_default_theme


class TestHierarchicalHelpFormatter:
  """Test HierarchicalHelpFormatter class functionality."""

  def setup_method(self):
    """Set up test formatter."""
    # Create minimal parser for testing
    self.parser = argparse.ArgumentParser(
      prog='test_cli',
      description='Test FreyjaCLI for formatter testing'
    )

    # Create formatter without theme initially
    self.formatter = HierarchicalHelpFormatter(
      prog='test_cli'
    )

  def test_formatter_initialization_no_theme(self):
    """Test formatter initialization without theme."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    assert formatter._theme is None
    assert formatter._color_formatter is None
    assert formatter._cmd_indent == 2
    assert formatter._arg_indent == 4
    assert formatter._desc_indent == 8

  def test_formatter_initialization_with_theme(self):
    """Test formatter initialization with theme."""
    theme = create_default_theme()
    formatter = HierarchicalHelpFormatter(prog='test_cli', theme=theme)

    assert formatter._theme == theme
    assert formatter._color_formatter is not None

  def test_console_width_detection(self):
    """Test console width detection and fallback."""
    with patch('os.get_terminal_size') as mock_get_size:
      # Test normal case
      mock_get_size.return_value = Mock(columns=120)
      formatter = HierarchicalHelpFormatter(prog='test_cli')
      assert formatter._console_width == 120

      # Test fallback to environment variable
      mock_get_size.side_effect = OSError()
      with patch.dict('os.environ', {'COLUMNS': '100'}):
        formatter = HierarchicalHelpFormatter(prog='test_cli')
        assert formatter._console_width == 100

      # Test default fallback
      mock_get_size.side_effect = OSError()
      with patch.dict('os.environ', {}, clear=True):
        formatter = HierarchicalHelpFormatter(prog='test_cli')
        assert formatter._console_width == 80

  def test_apply_style_no_theme(self):
    """Test _apply_style method without theme."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')
    result = formatter._apply_style("test text", "command_name")
    assert result == "test text"  # No styling applied

  def test_apply_style_with_theme(self):
    """Test _apply_style method with theme."""
    theme = create_default_theme()
    formatter = HierarchicalHelpFormatter(prog='test_cli', theme=theme)

    # Test that styling is applied (result should contain ANSI codes)
    result = formatter._apply_style("test text", "command_name")

    # Check if colors are enabled in the formatter
    if formatter._color_formatter and formatter._color_formatter.colors_enabled:
      assert result != "test text"  # Should be different due to ANSI codes
      assert "test text" in result  # Original text should be in result
    else:
      # If colors are disabled, result should be unchanged
      assert result == "test text"

  def test_get_display_width_plain_text(self):
    """Test _get_display_width with plain text."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')
    assert formatter._get_display_width("hello") == 5
    assert formatter._get_display_width("") == 0

  def test_get_display_width_with_ansi_codes(self):
    """Test _get_display_width strips ANSI codes correctly."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    # Text with ANSI color codes should report correct width
    ansi_text = "\x1b[32mhello\x1b[0m"  # Green "hello"
    assert formatter._get_display_width(ansi_text) == 5

    # More complex ANSI codes
    complex_ansi = "\x1b[1;32;48;5;231mhello world\x1b[0m"
    assert formatter._get_display_width(complex_ansi) == 11

  def test_wrap_text_basic(self):
    """Test _wrap_text method with basic text."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    text = "This is a test string that should be wrapped properly."
    lines = formatter._wrap_text(text, indent=4, width=40)

    assert len(lines) > 1  # Should wrap
    assert all(line.startswith("    ") for line in lines)  # All lines indented

  def test_wrap_text_empty(self):
    """Test _wrap_text with empty text."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')
    lines = formatter._wrap_text("", indent=4, width=80)
    assert lines == []

  def test_wrap_text_minimum_width(self):
    """Test _wrap_text respects minimum width."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    # Very small width should still use minimum
    text = "This is a test string."
    lines = formatter._wrap_text(text, indent=70, width=80)

    # Should still wrap despite small available width
    assert len(lines) >= 1

  def test_analyze_arguments_empty_parser(self):
    """Test _analyze_arguments with empty parser."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')
    required, optional = formatter._analyze_arguments(None)

    assert required == []
    assert optional == []

  def test_analyze_arguments_with_options(self):
    """Test _analyze_arguments with various option types."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    # Create parser with different argument types
    parser = argparse.ArgumentParser()
    parser.add_argument('--required-arg', required=True, help='Required argument')
    parser.add_argument('--optional-arg', help='Optional argument')
    parser.add_argument('--flag', action='store_true', help='Boolean flag')
    parser.add_argument('--with-metavar', metavar='VALUE', help='Arg with metavar')

    required, optional = formatter._analyze_arguments(parser)

    # Check required args (returned as list of tuples: (name, help))
    assert len(required) == 1
    required_names = [name for name, _ in required]
    assert '--required-arg REQUIRED_ARG' in required_names

    # Check optional args (should have 3: optional-arg, flag, with-metavar)
    assert len(optional) == 3

    # Find specific optional args
    optional_names = [name for name, _ in optional]
    assert '--optional-arg OPTIONAL_ARG' in optional_names
    assert '--flag' in optional_names
    assert '--with-metavar VALUE' in optional_names

  def test_format_inline_description_no_description(self):
    """Test _format_inline_description with no description."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    lines = formatter._format_inline_description(
      name="command",
      description="",
      name_indent=2,
      description_column=20,
      style_name="command_name",
      style_description="command_description"
    )

    assert len(lines) == 1
    assert lines[0] == "  command"

  def test_format_inline_description_with_colon(self):
    """Test _format_inline_description with colon."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    lines = formatter._format_inline_description(
      name="command",
      description="Test description",
      name_indent=2,
      description_column=0,  # Not used for colons
      style_name="command_name",
      style_description="command_description",
      add_colon=True
    )

    assert len(lines) == 1
    # Command descriptions now align at same column as options (not colon + 1 space)
    assert "command:" in lines[0]
    assert "Test description" in lines[0]

  def test_format_inline_description_wrapping(self):
    """Test _format_inline_description with long text that needs wrapping."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')
    formatter._console_width = 40  # Force small width for testing

    long_description = "This is a very long description that should definitely wrap to multiple lines when displayed in the help text."

    lines = formatter._format_inline_description(
      name="cmd",
      description=long_description,
      name_indent=2,
      description_column=10,
      style_name="command_name",
      style_description="command_description"
    )

    # Should wrap to multiple lines
    assert len(lines) > 1
    # First line should contain command name
    assert "cmd" in lines[0]

  def test_format_inline_description_alignment(self):
    """Test _format_inline_description column alignment."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    lines = formatter._format_inline_description(
      name="short",
      description="Description",
      name_indent=2,
      description_column=20,
      style_name="command_name",
      style_description="command_description"
    )

    assert len(lines) == 1
    line = lines[0]

    # Check that description starts at approximately the right column
    # (accounting for ANSI codes if theme is enabled)
    assert "short" in line
    assert "Description" in line

  def test_find_subparser_exists(self):
    """Test _find_subparser when subparser exists."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    # Create parser with subparsers
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    sub = subparsers.add_parser('test-cmd')

    # Find the subparser
    found = formatter._find_subparser(parser, 'test-cmd')
    assert found == sub

  def test_find_subparser_not_exists(self):
    """Test _find_subparser when subparser doesn't exist."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    subparsers.add_parser('existing-cmd')

    # Try to find non-existent subparser
    found = formatter._find_subparser(parser, 'nonexistent')
    assert found is None

  def test_find_subparser_no_subparsers(self):
    """Test _find_subparser with parser that has no subparsers."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    parser = argparse.ArgumentParser()
    found = formatter._find_subparser(parser, 'any-cmd')
    assert found is None


class TestHierarchicalFormatterWithCommandGroups:
  """Test HierarchicalHelpFormatter with CommandGroup descriptions."""

  def setup_method(self):
    """Set up test with mock parser that has CommandGroup description."""
    self.formatter = HierarchicalHelpFormatter(prog='test_cli')

    # Create mock parser with CommandGroup description
    self.parser = argparse.ArgumentParser()
    self.parser._command_group_description = "Custom group description from decorator"

  def test_format_group_with_command_group_description(self):
    """Test that CommandGroup descriptions are used in formatting."""
    # Create mock group parser
    group_parser = Mock()
    group_parser._command_group_description = "Database operations and management"
    group_parser._commands = {'create': 'Create database', 'migrate': 'Run migrations'}
    group_parser.description = "Default description"
    group_parser._actions = []  # Add empty actions list to avoid iteration error

    # Mock _find_subparser to return mock subparsers
    def mock_find_subparser(parser, name):
      mock_sub = Mock()
      mock_sub._actions = []  # Empty actions list
      return mock_sub

    self.formatter._find_subparser = mock_find_subparser

    # Mock other required methods
    self.formatter._calculate_group_dynamic_columns = Mock(return_value=(20, 30))
    self.formatter._format_command_with_args_global_command = Mock(return_value=['  cmd_group: description'])

    # Test the formatting
    lines = self.formatter._format_group_with_command_groups_global(
      name="db",
      parser=group_parser,
      base_indent=2,
      unified_cmd_desc_column=25,  # Add unified command description column
      global_option_column=40
    )

    # Should use the CommandGroup description in formatting
    assert len(lines) > 0
    # The first line should contain the formatted group name with description
    formatted_line = lines[0]
    # Command descriptions now align at column 25, not colon + 1 space
    assert "db:" in formatted_line
    assert "Database operations and management" in formatted_line

  def test_format_group_without_command_group_description(self):
    """Test formatting falls back to default when no CommandGroup description."""
    # Create mock group parser without CommandGroup description
    group_parser = Mock()
    # No _command_group_description attribute - set to None to avoid getattr returning Mock
    group_parser._command_group_description = None
    group_parser.description = "Default group description"
    group_parser.help = ""  # Ensure help is a string, not a Mock
    group_parser._commands = {}
    group_parser._actions = []  # Add empty actions list to avoid iteration error

    lines = self.formatter._format_group_with_command_groups_global(
      name="admin",
      parser=group_parser,
      base_indent=2,
      unified_cmd_desc_column=25,  # Add unified command description column
      global_option_column=40
    )

    # Should use default formatting
    assert len(lines) > 0


class TestHierarchicalFormatterIntegration:
  """Integration tests for HierarchicalHelpFormatter with actual parsers."""

  def test_format_action_with_subparsers(self):
    """Test _format_action with SubParsersAction."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    # Create parser with command groups
    parser = argparse.ArgumentParser(formatter_class=lambda *args, **kwargs: formatter)
    subparsers = parser.add_subparsers(dest='command')

    # Add a simple command group
    sub = subparsers.add_parser('test-cmd', help='Test command')
    sub.add_argument('--option', help='Test option')

    # Find the SubParsersAction
    subparsers_action = None
    for action in parser._actions:
      if isinstance(action, argparse._SubParsersAction):
        subparsers_action = action
        break

    assert subparsers_action is not None

    # Test formatting (should not raise exceptions)
    formatted = formatter._format_action(subparsers_action)
    assert isinstance(formatted, str)
    assert 'test-cmd' in formatted

  def test_format_global_option_aligned(self):
    """Test _format_global_option_aligned with actual arguments."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    # Create an argument action
    parser = argparse.ArgumentParser()
    parser.add_argument('--verbose', '-v', action='store_true', help='Enable verbose output')

    # Get the action
    verbose_action = None
    for action in parser._actions:
      if action.dest == 'verbose':
        verbose_action = action
        break

    assert verbose_action is not None

    # Mock the global column calculation
    formatter._ensure_global_column_calculated = Mock(return_value=30)

    # Test formatting
    formatted = formatter._format_global_option_aligned(verbose_action)
    assert isinstance(formatted, str)
    # Check for either --verbose or -v (argparse may prefer the short form)
    assert ('--verbose' in formatted or '-v' in formatted)
    assert 'Enable verbose output' in formatted

  def test_full_help_formatting_integration(self):
    """Test complete help formatting with real parser structure."""
    # Create a parser similar to what FreyjaCLI would create
    parser = argparse.ArgumentParser(
      prog='test_cli',
      description='Test FreyjaCLI Application',
      formatter_class=lambda *args, **kwargs: HierarchicalHelpFormatter(*args, **kwargs)
    )

    # Add global options
    parser.add_argument('--verbose', action='store_true', help='Enable verbose output')
    parser.add_argument('--config', metavar='FILE', help='Configuration file')

    # Add command groups
    subparsers = parser.add_subparsers(title='COMMANDS', dest='command')

    # Flat command
    hello_cmd = subparsers.add_parser('hello', help='Greet someone')
    hello_cmd.add_argument('--name', default='World', help='Name to greet')

    # Group command (simulate what FreyjaCLI creates)
    user_group = subparsers.add_parser('user', help='User management operations')
    user_group._command_type = 'group'
    user_group._commands = {'create': 'Create user', 'delete': 'Delete user'}

    # Test that help can be generated without errors
    help_text = parser.format_help()

    # Basic sanity checks
    assert 'Test FreyjaCLI Application' in help_text
    assert 'COMMANDS:' in help_text
    assert 'hello' in help_text
    assert 'user' in help_text
    assert '--verbose' in help_text
    assert '--config' in help_text


class TestHierarchicalFormatterEdgeCases:
  """Test edge cases and error handling for HierarchicalHelpFormatter."""

  def test_format_with_very_long_names(self):
    """Test formatting with very long command/option names."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')
    formatter._console_width = 40  # Small console for testing

    long_name = "very-long-command-name-that-exceeds-normal-length"
    long_description = "This is a very long description that should wrap properly even with long command names."

    lines = formatter._format_inline_description(
      name=long_name,
      description=long_description,
      name_indent=2,
      description_column=20,  # Will be exceeded by long name
      style_name="command_name",
      style_description="command_description"
    )

    # Should handle gracefully
    assert len(lines) >= 1
    assert long_name in lines[0]

  def test_format_with_empty_console_width(self):
    """Test behavior with minimal console width."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')
    formatter._console_width = 10  # Very small

    lines = formatter._format_inline_description(
      name="cmd",
      description="Description text",
      name_indent=2,
      description_column=15,
      style_name="command_name",
      style_description="command_description"
    )

    # Should still produce output
    assert len(lines) >= 1

  def test_analyze_arguments_with_complex_actions(self):
    """Test _analyze_arguments with complex argument configurations."""
    formatter = HierarchicalHelpFormatter(prog='test_cli')

    parser = argparse.ArgumentParser()

    # Add arguments with various configurations
    parser.add_argument('--choices', choices=['a', 'b', 'c'], help='Argument with choices')
    parser.add_argument('--count', type=int, action='append', help='Repeatable integer argument')
    parser.add_argument('--store-const', action='store_const', const='value', help='Store constant')

    required, optional = formatter._analyze_arguments(parser)

    # All should be optional since none marked as required
    assert len(required) == 0
    assert len(optional) == 3

    # Check that different action types are handled
    optional_names = [name for name, _ in optional]
    assert any('--choices' in name for name in optional_names)
    assert any('--count' in name for name in optional_names)
    assert any('--store-const' in name for name in optional_names)



from typing import Union


class MathUtil:
  EPSILON: float = 1e-6
  Numeric = Union[int, float]

  @classmethod
  def clamp(cls, value: float, min_val: float, max_val: float) -> float:
    """
    Clamp a value between min and max bounds and return it.
      :value: The value to clamp
      :min_val: Minimum allowed value
      :max_val: Maximum allowed value

    Examples:
        MathUtil.clamp(5, 0, 10) # 5
        MathUtil.clamp(-5, 0, 10) # 0
        MathUtil.clamp(15, 0, 10) # 10
    """
    return max(min_val, min(value, max_val))

  @classmethod
  def minmax_range(cls, args: [Numeric], negative_lower: bool = False) -> tuple[Numeric, Numeric]:
    lower, upper = cls.minmax(*args)

    return cls.safe_negative(lower, negative_lower), upper

  @classmethod
  def minmax(cls, *args: Numeric) -> tuple[Numeric, Numeric]:
    """
    Return the minimum and maximum of a dynamic number of arguments.
      :args: Variable number of int or float arguments

    Raises:
        ValueError: If no arguments are provided
    """
    if not args:
      raise ValueError("minmax() requires at least one argument")

    return min(args), max(args)

  @classmethod
  def safe_negative(cls, value: Numeric, neg: bool = True) -> Numeric:
    """
    Return the negative of a dynamic number only if neg is True.
    :param value: Value to check and convert
    :param neg: Whether to convert to negative or not
    """
    return -value if neg else value

  @classmethod
  def percent(cls, val: int | float, max_val: int | float) -> float:
    if max_val < cls.EPSILON: raise ValueError("max_val is too small")
    return val / float(max_val)



# Refactored FreyjaCLI class - Simplified coordinator role only
from __future__ import annotations

import os
import sys
from typing import *

from contextlib import contextmanager
from freyja.cli import ClassHandler, ExecutionCoordinator
from freyja.command import CommandDiscovery, CommandExecutor
from freyja.parser import CommandParser
from freyja.utils.output_capture import OutputCaptureConfig

Target = Union[Type[Any], Sequence[Type[Any]]]

class FreyjaCLI:
  """
  Simplified FreyjaCLI coordinator that orchestrates command discovery, parsing, and execution.
  """

  def __init__(self, target: Target, title: Optional[str] = None, method_filter: Optional[callable] = None,
               theme=None, alphabetize: bool = True, completion: bool = True, theme_tuner=False,
               # Output capture parameters
               capture_output: bool = False, capture_stdout: bool = True, capture_stderr: bool = False,
               capture_stdin: bool = False, output_capture_config: Optional[Dict[str, Any]] = None):
    """
    Initialize FreyjaCLI with target and configuration.

    :param target: Class or list of classes to generate FreyjaCLI from
    :param title: FreyjaCLI application title
    :param method_filter: Optional filter for methods (class mode)
    :param theme: Optional theme for colored output
    :param alphabetize: Whether to sort cmd_tree and options alphabetically
    :param completion: Whether to enable shell completion
    :param capture_output: Enable output capture (opt-in, default: False)
    :param capture_stdout: Capture stdout when output capture is enabled
    :param capture_stderr: Capture stderr when output capture is enabled
    :param capture_stdin: Capture stdin when output capture is enabled
    :param output_capture_config: Advanced OutputCapture configuration
    """

    # Initialize discovery service (replaces TargetAnalyzer)
    self.discovery = CommandDiscovery(target=target, method_filter=method_filter, completion=completion,
                                  theme_tuner=theme_tuner)

    # Get target mode and validation from discovery
    self.target_mode = self.discovery.mode

    # Validate class mode for command collisions when multiple classes are present
    if self.discovery.target_classes and len(self.discovery.target_classes) > 1:
      handler = ClassHandler()
      handler.validate_classes(self.discovery.target_classes)

    # Set title based on target
    self.title = title or self.discovery.generate_title()

    # Store only essential config
    self.theme = theme
    self.enable_completion = completion

    # Create output capture configuration
    self.output_capture_config = OutputCaptureConfig.from_kwargs(
      capture_output=capture_output,
      capture_stdout=capture_stdout,
      capture_stderr=capture_stderr,
      capture_stdin=capture_stdin,
      output_capture_config=output_capture_config
    )

    # Discover cmd_tree
    # self.discovery.cmd_tree = discovery.cmd_tree

    # Initialize command executors
    executors = self._initialize_executors()

    # Initialize execution coordinator
    self.execution_coordinator = ExecutionCoordinator(self.target_mode, executors, self.output_capture_config)
    # Pass command tree and CLI instance for completion
    self.execution_coordinator.command_tree = self.discovery.cmd_tree
    self.execution_coordinator.cli_instance = self

    # Initialize parser service
    self.parser_service = CommandParser(title=self.title, theme=theme, alphabetize=alphabetize, enable_completion=completion)

    # Set target_class and target_classes properties for compatibility
    self.target_class = self.discovery.primary_class
    self.target_classes = self.discovery.target_classes

    # Get command structure from discovery (no need to build separately)
    self.commands = self.discovery.cmd_tree.to_dict()

    # Essential compatibility properties only
    # Note: target_module removed - class-based CLI only

  def run(self, args: List[str] = None) -> Any:
    """
    Parse arguments and execute the appropriate command.

    :param args: Optional command line arguments (uses sys.argv if None)
    :return: Command execution result
    """
    result = None

    # Handle completion requests early
    if self.enable_completion and self._is_completion_request():
      self._handle_completion()
    else:
      # Check for no-color flag
      no_color = ExecutionCoordinator.check_no_color_flag(args or [])

      # Create parser and parse arguments
      parser = self.create_parser(no_color=no_color)

      # Parse and execute with context
      result = self._execute_with_context(parser, args)

    return result

  # Output Capture Public API
  @property
  def output_capture(self) -> Optional['OutputCapture']:
    """Access the OutputCapture instance if enabled."""
    return self.execution_coordinator.output_capture

  def get_captured_output(self, stream: str = 'stdout') -> Optional[str]:
    """Get captured output if capture is enabled.
    
    :param stream: Stream name ('stdout', 'stderr', 'stdin')
    :return: Captured content or None if capture disabled
    """
    if self.output_capture:
      return self.output_capture.get_output(stream)
    return None

  def get_all_captured_output(self) -> Dict[str, Optional[str]]:
    """Get all captured output if capture is enabled.
    
    :return: Dictionary with captured content for each stream
    """
    if self.output_capture:
      return self.output_capture.get_all_output()
    return {'stdout': None, 'stderr': None, 'stdin': None}

  def clear_captured_output(self) -> None:
    """Clear captured output buffer."""
    if self.output_capture:
      self.output_capture.clear()

  def enable_output_capture(self, **kwargs) -> None:
    """Enable output capture dynamically.
    
    :param kwargs: OutputCaptureConfig parameters (capture_stdout, capture_stderr, etc.)
    """
    self.execution_coordinator.enable_output_capture(**kwargs)

  def disable_output_capture(self) -> None:
    """Disable output capture dynamically."""
    self.execution_coordinator.disable_output_capture()

  @contextmanager
  def capture_output(self, **kwargs):
    """Context manager for temporary output capture.
    
    :param kwargs: OutputCaptureConfig parameters
    """
    was_enabled = self.execution_coordinator.has_output_capture()
    old_config = None
    
    if was_enabled:
      old_config = self.execution_coordinator.get_output_config()
    
    try:
      self.enable_output_capture(**kwargs)
      yield self.output_capture
    finally:
      if was_enabled and old_config:
        self.execution_coordinator.set_output_config(old_config)
      elif not was_enabled:
        self.disable_output_capture()

  def _execute_with_context(self, parser, args) -> Any:
    """Execute command with FreyjaCLI context."""
    # Add context information to the execution coordinator
    self.execution_coordinator.command_tree = self.discovery.cmd_tree
    self.execution_coordinator.cli_instance = self

    return self.execution_coordinator.parse_and_execute(parser, args)

  def _initialize_executors(self) -> dict:
    """Initialize command executors based on target mode."""
    # Create color formatter from theme
    color_formatter = None
    theme = self.theme
    if not theme:
      from freyja.theme.defaults import create_default_theme
      theme = create_default_theme()
    
    if theme:
      from freyja.theme import ColorFormatter
      color_formatter = ColorFormatter()
    
    if self.discovery.target_classes and len(self.discovery.target_classes) > 1:
      # Multiple classes: create executor for each class
      return {
        target_class: CommandExecutor(
          target_class=target_class,
          color_formatter=color_formatter,
          verbose=False  # Will be updated during execution
        )
        for target_class in self.discovery.target_classes
      }

    # Single class: create single primary executor
    return {
      'primary': CommandExecutor(
        target_class=self.discovery.primary_class,
        color_formatter=color_formatter,
        verbose=False  # Will be updated during execution
      )
    }

  def _is_completion_request(self) -> bool:
    """
    Check if this is a shell completion request.

    Uses multiple indicators to ensure we only enter completion mode
    when explicitly requested by shell completion systems.
    """
    # Check for main completion environment variable
    if os.getenv('_FREYJA_COMPLETE') is not None:
      return True

    # Check for shell-specific completion variables
    completion_vars = [
      '_FREYJA_COMPLETE_ZSH',
      '_FREYJA_COMPLETE_BASH',
      '_FREYJA_COMPLETE_FISH',
      '_FREYJA_COMPLETE_POWERSHELL'
    ]

    for var in completion_vars:
      if os.getenv(var) is not None:
        return True

    # Check for --_complete flag (legacy support)
    if '--_complete' in sys.argv:
      return True

    return False

  def _handle_completion(self):
    """
    Handle shell completion request.

    Ensures completion mode is completely isolated from normal execution.
    """
    try:
      # Use the execution coordinator's completion handler
      result = self.execution_coordinator._handle_completion_request()

      # Explicitly exit after completion to prevent any further processing
      sys.exit(0 if result == 0 else 1)

    except Exception as e:
      # In case of completion errors, fail gracefully
      print(f"Completion error: {e}", file=sys.stderr)
      sys.exit(1)
    finally:
      # Ensure we exit - completion should never continue to normal execution
      sys.exit(0)

  def create_parser(self, no_color: bool = False):
    """Create argument parser using pre-built command tree."""
    return self.parser_service.create_parser(
      command_tree=self.commands,
      target_mode=self.target_mode.value,
      target_class=self.target_class,
      no_color=no_color
    )



"""Complete color theme configuration with adjustment capabilities."""
from __future__ import annotations

from .enums import Back, Fore, ForeUniversal
from .rgb import RGB, AdjustStrategy
from .theme_style import CommandStyleSection, ThemeStyle


class Theme:
  """
  Complete color theme configuration for FreyjaCLI output with dynamic adjustment capabilities.
  Defines styling for all major UI elements in the help output with optional color adjustment.

  Uses hierarchical CommandStyleSection structure internally, with backward compatibility properties
  for existing flat attribute access patterns.
  """

  def __init__(self,
               # Hierarchical sections (new structure)
               topLevelCommandSection: CommandStyleSection | None = None,
               commandGroupSection: CommandStyleSection | None = None,
               groupedCommandSection: CommandStyleSection | None = None,
               # Non-sectioned attributes
               title: ThemeStyle | None = None,
               subtitle: ThemeStyle | None = None,
               required_asterisk: ThemeStyle | None = None,
               command_output: ThemeStyle | None = None,
               # Backward compatibility: flat attributes (legacy constructor support)
               command_name: ThemeStyle | None = None,
               command_description: ThemeStyle | None = None,
               command_group_name: ThemeStyle | None = None,
               command_group_description: ThemeStyle | None = None,
               grouped_command_name: ThemeStyle | None = None,
               grouped_command_description: ThemeStyle | None = None,
               option_name: ThemeStyle | None = None,
               option_description: ThemeStyle | None = None,
               command_group_option_name: ThemeStyle | None = None,
               command_group_option_description: ThemeStyle | None = None,
               grouped_command_option_name: ThemeStyle | None = None,
               grouped_command_option_description: ThemeStyle | None = None,
               # Adjustment settings
               adjust_strategy: AdjustStrategy = AdjustStrategy.LINEAR,
               adjust_percent: float = 0.0):
    """Initialize theme with hierarchical sections or backward compatible flat attributes."""
    if adjust_percent < -5.0 or adjust_percent > 5.0:
      raise ValueError(f"adjust_percent must be between -5.0 and 5.0, got {adjust_percent}")

    self.adjust_strategy = adjust_strategy
    self.adjust_percent = adjust_percent

    # Handle both hierarchical and flat initialization patterns
    if topLevelCommandSection is not None or commandGroupSection is not None or groupedCommandSection is not None:
      # New hierarchical initialization
      self.topLevelCommandSection = topLevelCommandSection or CommandStyleSection(
        command_name=ThemeStyle(), command_description=ThemeStyle(),
        option_name=ThemeStyle(), option_description=ThemeStyle()
      )
      self.commandGroupSection = commandGroupSection or CommandStyleSection(
        command_name=ThemeStyle(), command_description=ThemeStyle(),
        option_name=ThemeStyle(), option_description=ThemeStyle()
      )
      self.groupedCommandSection = groupedCommandSection or CommandStyleSection(
        command_name=ThemeStyle(), command_description=ThemeStyle(),
        option_name=ThemeStyle(), option_description=ThemeStyle()
      )
    else:
      # Legacy flat initialization - construct sections from flat attributes
      self.topLevelCommandSection = CommandStyleSection(
        command_name=command_name or ThemeStyle(),
        command_description=command_description or ThemeStyle(),
        option_name=option_name or ThemeStyle(),
        option_description=option_description or ThemeStyle()
      )
      self.commandGroupSection = CommandStyleSection(
        command_name=command_group_name or ThemeStyle(),
        command_description=command_group_description or ThemeStyle(),
        option_name=command_group_option_name or ThemeStyle(),
        option_description=command_group_option_description or ThemeStyle()
      )
      self.groupedCommandSection = CommandStyleSection(
        command_name=grouped_command_name or ThemeStyle(),
        command_description=grouped_command_description or ThemeStyle(),
        option_name=grouped_command_option_name or ThemeStyle(),
        option_description=grouped_command_option_description or ThemeStyle()
      )

    # Non-sectioned attributes
    self.title = title or ThemeStyle()
    self.subtitle = subtitle or ThemeStyle()
    self.required_asterisk = required_asterisk or ThemeStyle()
    self.command_output = command_output or ThemeStyle()

  # Backward compatibility properties for flat attribute access

  # Top-level command properties
  @property
  def command_name(self) -> ThemeStyle:
    """Top-level command name style (backward compatibility)."""
    return self.topLevelCommandSection.command_name

  @command_name.setter
  def command_name(self, value: ThemeStyle):
    self.topLevelCommandSection.command_name = value

  @property
  def command_description(self) -> ThemeStyle:
    """Top-level command description style (backward compatibility)."""
    return self.topLevelCommandSection.command_description

  @command_description.setter
  def command_description(self, value: ThemeStyle):
    self.topLevelCommandSection.command_description = value

  @property
  def option_name(self) -> ThemeStyle:
    """Top-level option name style (backward compatibility)."""
    return self.topLevelCommandSection.option_name

  @option_name.setter
  def option_name(self, value: ThemeStyle):
    self.topLevelCommandSection.option_name = value

  @property
  def option_description(self) -> ThemeStyle:
    """Top-level option description style (backward compatibility)."""
    return self.topLevelCommandSection.option_description

  @option_description.setter
  def option_description(self, value: ThemeStyle):
    self.topLevelCommandSection.option_description = value

  # Command group properties
  @property
  def command_group_name(self) -> ThemeStyle:
    """Command group name style (backward compatibility)."""
    return self.commandGroupSection.command_name

  @command_group_name.setter
  def command_group_name(self, value: ThemeStyle):
    self.commandGroupSection.command_name = value

  @property
  def command_group_description(self) -> ThemeStyle:
    """Command group description style (backward compatibility)."""
    return self.commandGroupSection.command_description

  @command_group_description.setter
  def command_group_description(self, value: ThemeStyle):
    self.commandGroupSection.command_description = value

  @property
  def command_group_option_name(self) -> ThemeStyle:
    """Command group option name style (backward compatibility)."""
    return self.commandGroupSection.option_name

  @command_group_option_name.setter
  def command_group_option_name(self, value: ThemeStyle):
    self.commandGroupSection.option_name = value

  @property
  def command_group_option_description(self) -> ThemeStyle:
    """Command group option description style (backward compatibility)."""
    return self.commandGroupSection.option_description

  @command_group_option_description.setter
  def command_group_option_description(self, value: ThemeStyle):
    self.commandGroupSection.option_description = value

  # Grouped command properties
  @property
  def grouped_command_name(self) -> ThemeStyle:
    """Grouped command name style (backward compatibility)."""
    return self.groupedCommandSection.command_name

  @grouped_command_name.setter
  def grouped_command_name(self, value: ThemeStyle):
    self.groupedCommandSection.command_name = value

  @property
  def grouped_command_description(self) -> ThemeStyle:
    """Grouped command description style (backward compatibility)."""
    return self.groupedCommandSection.command_description

  @grouped_command_description.setter
  def grouped_command_description(self, value: ThemeStyle):
    self.groupedCommandSection.command_description = value

  @property
  def grouped_command_option_name(self) -> ThemeStyle:
    """Grouped command option name style (backward compatibility)."""
    return self.groupedCommandSection.option_name

  @grouped_command_option_name.setter
  def grouped_command_option_name(self, value: ThemeStyle):
    self.groupedCommandSection.option_name = value

  @property
  def grouped_command_option_description(self) -> ThemeStyle:
    """Grouped command option description style (backward compatibility)."""
    return self.groupedCommandSection.option_description

  @grouped_command_option_description.setter
  def grouped_command_option_description(self, value: ThemeStyle):
    self.groupedCommandSection.option_description = value

  def create_adjusted_copy(self, adjust_percent: float, adjust_strategy: AdjustStrategy | None = None) -> Theme:
    """Create a new theme with adjusted colors.

    :param adjust_percent: Adjustment percentage (-5.0 to 5.0)
    :param adjust_strategy: Optional strategy override
    :return: New Theme instance with adjusted colors
    """
    if adjust_percent < -5.0 or adjust_percent > 5.0:
      raise ValueError(f"adjust_percent must be between -5.0 and 5.0, got {adjust_percent}")

    strategy = adjust_strategy or self.adjust_strategy

    # Temporarily set adjustment parameters for the adjustment process
    old_adjust_percent = self.adjust_percent
    old_adjust_strategy = self.adjust_strategy
    self.adjust_percent = adjust_percent
    self.adjust_strategy = strategy

    try:
      # Create adjusted CommandStyleSection instances
      adjusted_top_level = CommandStyleSection(
        command_name=self.get_adjusted_style(self.topLevelCommandSection.command_name),
        command_description=self.get_adjusted_style(self.topLevelCommandSection.command_description),
        option_name=self.get_adjusted_style(self.topLevelCommandSection.option_name),
        option_description=self.get_adjusted_style(self.topLevelCommandSection.option_description)
      )

      adjusted_command_group = CommandStyleSection(
        command_name=self.get_adjusted_style(self.commandGroupSection.command_name),
        command_description=self.get_adjusted_style(self.commandGroupSection.command_description),
        option_name=self.get_adjusted_style(self.commandGroupSection.option_name),
        option_description=self.get_adjusted_style(self.commandGroupSection.option_description)
      )

      adjusted_grouped_command = CommandStyleSection(
        command_name=self.get_adjusted_style(self.groupedCommandSection.command_name),
        command_description=self.get_adjusted_style(self.groupedCommandSection.command_description),
        option_name=self.get_adjusted_style(self.groupedCommandSection.option_name),
        option_description=self.get_adjusted_style(self.groupedCommandSection.option_description)
      )

      # Create new theme using hierarchical constructor
      new_theme = Theme(
        topLevelCommandSection=adjusted_top_level,
        commandGroupSection=adjusted_command_group,
        groupedCommandSection=adjusted_grouped_command,
        title=self.get_adjusted_style(self.title),
        subtitle=self.get_adjusted_style(self.subtitle),
        required_asterisk=self.get_adjusted_style(self.required_asterisk),
        adjust_strategy=strategy,
        adjust_percent=adjust_percent
      )
    finally:
      # Restore original adjustment parameters
      self.adjust_percent = old_adjust_percent
      self.adjust_strategy = old_adjust_strategy

    return new_theme

  def get_adjusted_style(self, original: ThemeStyle) -> ThemeStyle:
    """Create a ThemeStyle with adjusted colors.

    :param original: Original ThemeStyle
    :return: ThemeStyle with adjusted colors
    """
    adjusted_fg = None
    if original.fg:
      # Use RGB adjustment method directly
      adjusted_fg = original.fg.adjust(brightness=self.adjust_percent, strategy=self.adjust_strategy)

    # Background adjustment disabled for now (as in original)
    adjusted_bg = original.bg

    return ThemeStyle(
      fg=adjusted_fg, bg=adjusted_bg, bold=original.bold, italic=original.italic, dim=original.dim,
      underline=original.underline
    )


def create_default_theme() -> Theme:
  """Create a default color theme using universal colors for optimal cross-platform compatibility."""
  # Create hierarchical sections
  top_level_section = CommandStyleSection(
    command_name=ThemeStyle(bold=True),
    command_description=ThemeStyle(bold=True, italic=True,),
    option_name=ThemeStyle(),
    option_description=ThemeStyle(fg=RGB.from_rgb(ForeUniversal.BROWN.value), bold=True)
  )

  command_group_section = CommandStyleSection(
    command_name=ThemeStyle(bold=True),
    command_description=ThemeStyle(fg=RGB.from_rgb(ForeUniversal.BROWN.value), bold=True),
    option_name=ThemeStyle(),
    option_description=ThemeStyle(fg=RGB.from_rgb(ForeUniversal.BROWN.value), bold=True)
  )

  grouped_command_section = CommandStyleSection(
    command_name=ThemeStyle(bold=True),
    command_description=ThemeStyle(fg=RGB.from_rgb(ForeUniversal.BROWN.value), bold=True, italic=True),
    option_name=ThemeStyle(),
    option_description=ThemeStyle(fg=RGB.from_rgb(ForeUniversal.BROWN.value))
  )

  return Theme(
    topLevelCommandSection=top_level_section,
    commandGroupSection=command_group_section,
    groupedCommandSection=grouped_command_section,
    title=ThemeStyle(fg=RGB.from_rgb(ForeUniversal.TEAL.value), bold=True),
    subtitle=ThemeStyle(fg=RGB.from_rgb(ForeUniversal.TEAL.value), italic=True),
    required_asterisk=ThemeStyle(fg=RGB.from_rgb(ForeUniversal.GOLD.value)),
    adjust_percent=0.0
  )


def create_default_theme_colorful() -> Theme:
  """Create a colorful theme with traditional terminal colors."""
  # Create hierarchical sections
  top_level_section = CommandStyleSection(
    command_name=ThemeStyle(fg=RGB.from_rgb(Fore.CYAN.value), bold=True),
    command_description=ThemeStyle(fg=RGB.from_rgb(Fore.LIGHTRED_EX.value)),
    option_name=ThemeStyle(fg=RGB.from_rgb(Fore.GREEN.value)),
    option_description=ThemeStyle(fg=RGB.from_rgb(Fore.YELLOW.value))
  )

  command_group_section = CommandStyleSection(
    command_name=ThemeStyle(fg=RGB.from_rgb(Fore.CYAN.value), bold=True),
    command_description=ThemeStyle(fg=RGB.from_rgb(Fore.LIGHTRED_EX.value)),
    option_name=ThemeStyle(fg=RGB.from_rgb(Fore.GREEN.value)),
    option_description=ThemeStyle(fg=RGB.from_rgb(Fore.YELLOW.value))
  )

  grouped_command_section = CommandStyleSection(
    command_name=ThemeStyle(fg=RGB.from_rgb(Fore.CYAN.value), italic=True, bold=True),
    command_description=ThemeStyle(fg=RGB.from_rgb(Fore.LIGHTRED_EX.value)),
    option_name=ThemeStyle(fg=RGB.from_rgb(Fore.GREEN.value)),
    option_description=ThemeStyle(fg=RGB.from_rgb(Fore.YELLOW.value))
  )

  return Theme(
    topLevelCommandSection=top_level_section,
    commandGroupSection=command_group_section,
    groupedCommandSection=grouped_command_section,
    title=ThemeStyle(fg=RGB.from_rgb(Fore.MAGENTA.value), bg=RGB.from_rgb(Back.LIGHTWHITE_EX.value), bold=True),
    subtitle=ThemeStyle(fg=RGB.from_rgb(Fore.YELLOW.value), italic=True),
    required_asterisk=ThemeStyle(fg=RGB.from_rgb(Fore.YELLOW.value))
  )


def create_no_color_theme() -> Theme:
  """Create a theme with no colors (fallback for non-color terminals)."""
  # Create hierarchical sections with no colors
  top_level_section = CommandStyleSection(
    command_name=ThemeStyle(),
    command_description=ThemeStyle(),
    option_name=ThemeStyle(),
    option_description=ThemeStyle()
  )

  command_group_section = CommandStyleSection(
    command_name=ThemeStyle(),
    command_description=ThemeStyle(),
    option_name=ThemeStyle(),
    option_description=ThemeStyle()
  )

  grouped_command_section = CommandStyleSection(
    command_name=ThemeStyle(),
    command_description=ThemeStyle(),
    option_name=ThemeStyle(),
    option_description=ThemeStyle()
  )

  return Theme(
    topLevelCommandSection=top_level_section,
    commandGroupSection=command_group_section,
    groupedCommandSection=grouped_command_section,
    title=ThemeStyle(),
    subtitle=ThemeStyle(),
    required_asterisk=ThemeStyle()
  )



"""Completion installation and management."""

import os
import sys
from pathlib import Path

from .base import CompletionHandler


class CompletionInstaller:
  """Handles installation of shell completion scripts."""

  def __init__(self, handler: CompletionHandler, prog_name: str, command_patterns: list | None = None):
    """Initialize installer with handler and program name.

    :param handler: Completion handler for specific shell
    :param prog_name: Name of the program to install completion for (will be normalized to basename)
    :param command_patterns: Additional command patterns to register completion for
    """
    from pathlib import Path

    self.handler = handler
    # Normalize program name to basename to handle path-based invocations
    # This ensures examples/cls_example becomes cls_example for consistent completion registration
    self.prog_name = Path(prog_name).name
    self.command_patterns = command_patterns or []
    self.shell = handler.detect_shell()

  def install(self, shell: str | None = None, force: bool = False) -> bool:
    """Install completion for specified or detected shell.

    :param shell: Target shell (auto-detect if None)
    :param force: Force overwrite existing completion
    :return: True if installation successful
    """
    target_shell = shell or self.shell

    if not target_shell:
      print("Could not detect shell. Please specify shell manually.", file=sys.stderr)
      return False

    if target_shell == 'bash':
      return self._install_bash_completion(force)
    elif target_shell == 'zsh':
      return self._install_zsh_completion(force)
    elif target_shell == 'fish':
      return self._install_fish_completion(force)
    elif target_shell == 'powershell':
      return self._install_powershell_completion(force)
    else:
      print(f"Unsupported shell: {target_shell}", file=sys.stderr)
      return False

  def _install_bash_completion(self, force: bool) -> bool:
    """Install bash completion."""
    # Try user completion directory first
    completion_dir = Path.home() / '.bash_completion.d'
    if not completion_dir.exists():
      completion_dir.mkdir(parents=True, exist_ok=True)

    completion_file = completion_dir / f'{self.prog_name}_completion'

    # Check if already exists
    if completion_file.exists() and not force:
      print(f"Completion already exists at {completion_file}. Use --force to overwrite.")
      return False

    # Generate and write completion script
    script = self.handler.generate_script(self.prog_name)
    completion_file.write_text(script)

    # Add sourcing to .bashrc if not already present
    bashrc = Path.home() / '.bashrc'
    source_line = f'source "{completion_file}"'

    if bashrc.exists():
      bashrc_content = bashrc.read_text()
      if source_line not in bashrc_content:
        with open(bashrc, 'a') as f:
          f.write(f'\n# Freyja completion for {self.prog_name}\n')
          f.write(f'{source_line}\n')
        print(f"Added completion sourcing to {bashrc}")

    print(f"Bash completion installed to {completion_file}")
    print("Restart your shell or run: source ~/.bashrc")
    return True

  def _install_zsh_completion(self, force: bool) -> bool:
    """Install zsh completion."""
    # Try standard user completion directories in order of preference
    completion_dirs = [
      Path.home() / '.zfunc',  # Most standard location
      Path.home() / '.zsh' / 'completions',
    ]

    # Use first directory that exists or can be created
    completion_dir = None
    for dir_path in completion_dirs:
      if dir_path.exists():
        completion_dir = dir_path
        break
      try:
        dir_path.mkdir(parents=True, exist_ok=True)
        completion_dir = dir_path
        break
      except:
        continue

    if not completion_dir:
      completion_dir = completion_dirs[0]  # Default to .zfunc
    completion_dir.mkdir(parents=True, exist_ok=True)
    completion_file = completion_dir / f'_{self.prog_name}'

    # Check if already exists
    if completion_file.exists() and not force:
      print(f"Completion already exists at {completion_file}. Use --force to overwrite.")
      return False

    # Generate and write completion script with command patterns
    script = self.handler.generate_script(self.prog_name, self.command_patterns)
    completion_file.write_text(script)

    print(f"Zsh completion installed to {completion_file}")

    # Clear completion cache to ensure new completions are loaded
    zcompdump_files = [
      Path.home() / '.zcompdump',
      Path.home() / f'.zcompdump-{os.getenv("HOST", "")}-{os.getenv("ZSH_VERSION", "")}',
    ]

    for zcompdump_file in zcompdump_files:
      if zcompdump_file.exists():
        try:
          zcompdump_file.unlink()
          print(f"Cleared completion cache: {zcompdump_file}")
        except:
          pass  # Ignore if we can't remove it

    # Check if completion directory is in fpath and auto-configure if needed
    import subprocess
    needs_fpath_config = True
    try:
      result = subprocess.run(['zsh', '-c', 'echo $fpath'], capture_output=True, text=True)
      if completion_dir.as_posix() in result.stdout:
        needs_fpath_config = False
    except:
      pass  # Assume we need to configure fpath

    if needs_fpath_config:
      # Try to automatically add fpath configuration to ~/.zshrc
      zshrc_path = Path.home() / '.zshrc'
      fpath_line = f'fpath=({completion_dir} $fpath)'

      try:
        if zshrc_path.exists():
          zshrc_content = zshrc_path.read_text()
          # Check if fpath line already exists in some form
          if str(completion_dir) not in zshrc_content or 'fpath=' not in zshrc_content:
            with open(zshrc_path, 'a') as f:
              f.write('\n# Freyja completion directory\n')
              f.write(f'{fpath_line}\n')
            print(f"Added fpath configuration to {zshrc_path}")
          else:
            print(f"fpath already configured in {zshrc_path}")
        else:
          # Create .zshrc with fpath configuration
          with open(zshrc_path, 'w') as f:
            f.write('# Freyja completion directory\n')
            f.write(f'{fpath_line}\n')
          print(f"Created {zshrc_path} with fpath configuration")

        print("Restart your shell or run: source ~/.zshrc && autoload -U compinit && compinit")
      except Exception:
        print("\nCould not automatically configure fpath. Please add this line to your ~/.zshrc:")
        print(f"{fpath_line}")
        print("Then run: autoload -U compinit && compinit")
    else:
      print("fpath already configured. Run: autoload -U compinit && compinit")

    return True

  def _install_fish_completion(self, force: bool) -> bool:
    """Install fish completion."""
    completion_dir = Path.home() / '.config' / 'fish' / 'completions'
    completion_dir.mkdir(parents=True, exist_ok=True)

    completion_file = completion_dir / f'{self.prog_name}.fish'

    # Check if already exists
    if completion_file.exists() and not force:
      print(f"Completion already exists at {completion_file}. Use --force to overwrite.")
      return False

    # Generate and write completion script
    script = self.handler.generate_script(self.prog_name)
    completion_file.write_text(script)

    print(f"Fish completion installed to {completion_file}")
    print("Restart your shell for changes to take effect")
    return True

  def _install_powershell_completion(self, force: bool) -> bool:
    """Install PowerShell completion."""
    # PowerShell profile path
    if os.name == 'nt':
      # Windows PowerShell
      profile_dir = Path(os.environ.get('USERPROFILE', '')) / 'Documents' / 'WindowsPowerShell'
    else:
      # PowerShell Core on Unix
      profile_dir = Path.home() / '.config' / 'powershell'

    profile_dir.mkdir(parents=True, exist_ok=True)
    profile_file = profile_dir / 'Microsoft.PowerShell_profile.ps1'

    # Generate completion script
    script = self.handler.generate_script(self.prog_name)

    # Check if profile exists and has our completion
    completion_marker = f'# Freyja completion for {self.prog_name}'

    if profile_file.exists():
      profile_content = profile_file.read_text()
      if completion_marker in profile_content and not force:
        print(f"Completion already installed in {profile_file}. Use --force to overwrite.")
        return False

      # Remove old completion if forcing
      if force and completion_marker in profile_content:
        lines = profile_content.split('\n')
        new_lines = []
        skip_next = False

        for line in lines:
          if completion_marker in line:
            skip_next = True
            continue
          if skip_next and line.strip().startswith('Register-ArgumentCompleter'):
            skip_next = False
            continue
          new_lines.append(line)

        profile_content = '\n'.join(new_lines)
    else:
      profile_content = ''

    # Add completion to profile
    with open(profile_file, 'w') as f:
      f.write(profile_content)
      f.write(f'\n{completion_marker}\n')
      f.write(script)

    print(f"PowerShell completion installed to {profile_file}")
    print("Restart PowerShell for changes to take effect")
    return True

  def uninstall(self, shell: str | None = None) -> bool:
    """Remove installed completion.

    :param shell: Target shell (auto-detect if None)
    :return: True if uninstallation successful
    """
    target_shell = shell or self.shell

    if not target_shell:
      print("Could not detect shell. Please specify shell manually.", file=sys.stderr)
      return False

    success = False

    if target_shell == 'bash':
      completion_file = Path.home() / '.bash_completion.d' / f'{self.prog_name}_completion'
      if completion_file.exists():
        completion_file.unlink()
        success = True

    elif target_shell == 'zsh':
      completion_dirs = [
        Path.home() / '.zsh' / 'completions',
        Path.home() / '.oh-my-zsh' / 'completions',
        Path('/usr/local/share/zsh/site-functions')
      ]

      for dir_path in completion_dirs:
        completion_file = dir_path / f'_{self.prog_name}'
        if completion_file.exists():
          completion_file.unlink()
          success = True

    elif target_shell == 'fish':
      completion_file = Path.home() / '.config' / 'fish' / 'completions' / f'{self.prog_name}.fish'
      if completion_file.exists():
        completion_file.unlink()
        success = True

    elif target_shell == 'powershell':
      # For PowerShell, we would need to edit the profile file
      print("PowerShell completion uninstall requires manual removal from profile.")
      return False

    if success:
      print(f"Completion uninstalled for {target_shell}")
    else:
      print(f"No completion found to uninstall for {target_shell}")

    return success
